/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include "cgi.h"
#include "pairarray.h"
#include "strutil.h"
#include "memutil.h"
#include "logger.h"
#include "debug.h"

static request_t *
cgi_request_parse_get(void);
static request_t *
cgi_request_parse_post(void);
static request_t *
cgi_request_parse_post_normal(void);
static request_t *
cgi_request_parse_post_multipart(const char *boundary);
static int
cgi_request_parse_post_multipart_one(const char *src, int len, request_t *request);
static pairarray_t *
cgi_request_parse_header_content_disposition(const char *src);
static request_t *
cgi_request_parse_core(const char *data);
static int
cgi_request_parse_one(request_t *request, const char *data);

extern logger_t *glogger;

request_t *
cgi_request_get(void)
{
	request_t *dst = NULL;
	char *method;

	logger_entry_function(glogger);

	method = getenv("REQUEST_METHOD");

	logger_printf(glogger, LOGGER_DEBUG, "method: %s", method);

	if(strutil_isequal(method, "GET"))
	{
		dst = cgi_request_parse_get();
	}
	else if(strutil_isequal(method, "POST"))
	{
		dst = cgi_request_parse_post();
	}

	if(!dst)
	{
		goto ERROR;
	}

	logger_leave_function(glogger);

	return dst;

ERROR:
	logger_leave_function(glogger);

	return NULL;
}

static request_t *
cgi_request_parse_get(void)
{
	request_t *dst;

	logger_entry_function(glogger);

	dst = cgi_request_parse_core(getenv("QUERY_STRING"));

	if(!dst)
	{
		goto ERROR;
	}

	logger_printf(glogger, LOGGER_DEBUG, "success");

	logger_leave_function(glogger);

	return dst;

ERROR:
	logger_printf(glogger, LOGGER_DEBUG, "failure");

	logger_leave_function(glogger);

	return NULL;
}

static request_t *
cgi_request_parse_post(void)
{
	strarray_t * content_type;
	char *media_type;
	request_t *dst;

	content_type = strutil_split(getenv("CONTENT_TYPE"), ";", 0);

	if(!content_type || strarray_len(content_type) <= 0)
	{
		goto ERROR;
	}

	media_type = strutil_trim(strarray_get(content_type, 0));

	if(strutil_isequal(media_type, "application/x-www-form-urlencoded"))
	{
		dst = cgi_request_parse_post_normal();
	}
	else if(strutil_isequal(media_type, "multipart/form-data"))
	{
		int len, i;
		strarray_t *tmp;
		char *name, *value, *boundary = NULL;

		len = strarray_len(content_type);

		for(i = 1; i < len; i++)
		{
			tmp = strutil_split(strarray_get(content_type, i), "=", 1);

			name  = strutil_trim(strarray_get(tmp, 0));
			value = strutil_trim(strarray_get(tmp, 1));

			if(strutil_isequal(name, "boundary"))
			{
				boundary = strutil_unit("--", value, NULL);

				free(name); free(value);

				strarray_destroy(tmp);

				break;
			}

			free(name); free(value);

			strarray_destroy(tmp);
		}

		dst = cgi_request_parse_post_multipart(boundary);

		free(boundary);
	}
	else
	{
		goto ERROR;
	}

	free(media_type);

	strarray_destroy(content_type);

	return dst;

ERROR:
	free(media_type);

	strarray_destroy(content_type);

	return NULL;
}

static request_t *
cgi_request_parse_post_normal(void)
{
	request_t *dst;
	char *tmp;
	int len;

	len = strtol(getenv("CONTENT_LENGTH"), NULL, 10);

	tmp = (char *)calloc(len + 10000, sizeof(char));

	if(!tmp)
	{
		goto ERROR;
	}

	fread(tmp, sizeof(char), len, stdin);

	dst = cgi_request_parse_core(tmp);

	free(tmp);

	return dst;

ERROR:
	return NULL;
}

static request_t *
cgi_request_parse_post_multipart(const char *boundary)
{
	request_t *dst;
	int len;
	char *tmp = NULL, *sp, *ep;

	if(strutil_isempty(boundary))
	{
		goto ERROR;
	}

	dst = request_create();

	len = strtol(getenv("CONTENT_LENGTH"), NULL, 10);

	tmp = (char *)calloc(len + 1, sizeof(char));

	fread(tmp, sizeof(char), len, stdin);

	sp = tmp;

	ep = memutil_nstr(tmp, boundary, (tmp + len) - sp);

	if(!ep)
	{
		goto ERROR;
	}

	while(1)
	{
		sp = ep + strlen(boundary);

		if(sp[0] == '-' && sp[1] == '-')
		{
			break;
		}

		sp += 2;

		ep = memutil_nstr(sp, boundary, (tmp + len) - sp);

		if(!ep)
		{
			goto ERROR;
		}

		cgi_request_parse_post_multipart_one(sp, (ep - 2) - sp, dst);
	}

	free(tmp);

	return dst;

ERROR:
	free(tmp);

	request_destroy(dst);

	return NULL;
}

static int
cgi_request_parse_post_multipart_one(const char *src, int len, request_t *request)
{
	char *sp, *ep, *tmp, *name, *value, *filename;
	strarray_t *header;
	pairarray_t *content_disposition = NULL;

	sp = (char *)src;

	if(!sp)
	{
		goto ERROR;
	}

	while((ep = strstr(sp, "\r\n")) != NULL)
	{
		if(ep >= src + len)
		{
			break;
		}

		if(ep == sp)
		{
			ep += 2;

			if(!content_disposition || !strutil_isequal(pairarray_get_value_by_name(content_disposition, "disposition-type"), "form-data"))
			{
				break;
			}

			filename = pairarray_get_value_by_name(content_disposition, "filename");

			if(strutil_isempty(filename))
			{
				value = strutil_ndup(ep, len - (ep - src));

				request_append_item(request, pairarray_get_value_by_name(content_disposition, "name"), value);

				free(value);
			}
			else if(!strutil_isequal(filename, "\"\""))
			{
				request_upload_item_t *item;

				item = request_append_upload_item(request, pairarray_get_value_by_name(content_disposition, "name"));

				if(item)
				{
					int fd;

					fd = open(request_upload_item_get_path(item), O_WRONLY, 0666);

					if(fd > 0)
					{
						write(fd, ep, len - (ep - src));
					}

					close(fd);
				}
			}

			free(filename);

			pairarray_destroy(content_disposition);

			return 0;
		}

		tmp = strutil_ndup(sp, ep - sp);

		header = strutil_split(tmp, ":", 1);

		free(tmp);

		if(strarray_len(header) == 2)
		{
			name  = strutil_trim(strarray_get(header, 0));
			value = strutil_trim(strarray_get(header, 1));

			if(strutil_isequal(name, "content-disposition"))
			{
				content_disposition = cgi_request_parse_header_content_disposition(value);
			}

			free(name); free(value);
		}

		strarray_destroy(header);

		sp = ep + 2;
	}

	pairarray_destroy(content_disposition);

ERROR:
	return -1;
}

static pairarray_t *
cgi_request_parse_header_content_disposition(const char *src)
{
	strarray_t *tmp, *param;
	pairarray_t *dst;
	char *type, *name, *value;
	int len, i;

	tmp = strutil_split(src, ";", 0);

	if(strarray_len(tmp) <= 0)
	{
		goto ERROR;
	}

	dst = pairarray_create();

	type = strutil_trim(strarray_get(tmp, 0));

	pairarray_set_value_by_name(dst, "disposition-type", type);

	free(type);

	len = strarray_len(tmp);

	for(i = 1; i < len; i++)
	{
		param = strutil_split(strarray_get(tmp, i), "=", 1);

		name  = strutil_trim(strarray_get(param, 0));
		value = strutil_trim(strarray_get(param, 1));

		if(value)
		{
			int vlen;

			vlen = strlen(value);

			if(value[0] == '"' && value[vlen - 1] == '"')
			{
				if(vlen - 2)
				{
					memmove(value, value + 1, vlen - 2);
					value[vlen - 2] = '\0';
				}
			}
		}

		pairarray_set_value_by_name(dst, name, value);

		free(name); free(value);

		strarray_destroy(param);
	}

	strarray_destroy(tmp);

	return dst;

ERROR:
	return NULL;
}

static request_t *
cgi_request_parse_core(const char *data)
{
	request_t *dst;
	strarray_t *tmp;
	int len, i;

	logger_entry_function(glogger);

	dst = request_create();

	tmp = strutil_split(data, "&", 0);

	len = strarray_len(tmp);

	for(i = 0; i < len; i++)
	{
		cgi_request_parse_one(dst, strarray_get(tmp, i));
	}

	strarray_destroy(tmp);

	logger_leave_function(glogger);

	return dst;
}

static int
cgi_request_parse_one(request_t *request, const char *data)
{
	strarray_t *tmp;
	char *name, *value;

	logger_entry_function(glogger);

	if(!request || !data)
	{
		goto ERROR;
	}

	logger_printf(glogger, LOGGER_DEBUG, "data: %s", data);

	tmp = strutil_split(data, "=", 1);

	name  = strutil_trim(strarray_get(tmp, 0));

	value = strutil_trim(strarray_get(tmp, 1));

	logger_printf(glogger, LOGGER_DEBUG, "name: %s", name);
	logger_printf(glogger, LOGGER_DEBUG, "value: %s", value);

	request_append_item(request, name, value);

	free(name); free(value);

	strarray_destroy(tmp);

	logger_leave_function(glogger);

	return  0;

ERROR:
	logger_leave_function(glogger);

	return -1;
}
