/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "conf.h"
#include "strutil.h"
#include "strarray.h"
#include "debug.h"

struct conf
{
	list_t *sections;
};

struct conf_section
{
	char *name;
	list_t *items;
};

struct conf_section_item
{
	char *name;
	char *value;
};

static conf_section_t *
conf_section_create(void);
static void
conf_section_destroy(conf_section_t *section);
static conf_section_item_t *
conf_section_item_create(void);
static void
conf_section_item_destroy(conf_section_item_t *item);

conf_t *
conf_create(void)
{
	conf_t *conf;

	conf = (conf_t *)malloc(sizeof(conf_t));

	if(conf)
	{
		conf->sections = list_create();
	}

	return conf;
}

void
conf_destroy(conf_t *conf)
{
	if(conf)
	{
		while(list_size(conf->sections))
		{
			conf_section_destroy(list_remove(conf->sections, 0));
		}

		list_destroy(conf->sections);
	}

	free(conf);
}

int
conf_numberof_sections(const conf_t *conf)
{
	return conf ? list_size(conf->sections) : -1;
}

int
conf_load(conf_t *conf, const char *file)
{
	FILE *fp;
	char buf[4096], *section_name, *name, *value;
	conf_section_t *section = NULL;
	strarray_t *array;

	fp = fopen(file, "r");

	if(!fp)
	{
		goto ERROR;
	}

	while(fgets(buf, sizeof(buf), fp))
	{
		strutil_trim_direct(buf);

		if(strutil_isempty(buf) || buf[0] == '#')
		{
			continue;
		}

		if(buf[0] == '[' && buf[strlen(buf) - 1] == ']')
		{
			section_name = strutil_ndup(buf + 1, strlen(buf) - 2);

			if(!strutil_isempty(section_name))
			{
				section = conf_append_section(conf, section_name);
			}

			free(section_name);

			continue;
		}

		if(section)
		{
			array = strutil_split(buf, "=", 1);

			if(strarray_len(array) == 2)
			{
				name  = strutil_trim_direct(strarray_get(array, 0));
				value = strutil_trim_direct(strarray_get(array, 1));

				conf_section_append_item(section, name, value);
			}

			strarray_destroy(array);
		}
	}

	fclose(fp);

	return  0;

ERROR:
	return -1;
}

int
conf_save(const conf_t *conf, const char *file)
{
	FILE *fp;
	int num_sections, i, num_items, j;
	conf_section_t *section;
	conf_section_item_t *item;

	fp = fopen(file, "w");

	if(!fp)
	{
		goto ERROR;
	}

	num_sections = conf_numberof_sections(conf);

	for(i = 0; i < num_sections; i++)
	{
		section = conf_get_section(conf, i);

		fprintf(fp, "[%s]\n", conf_section_get_name(section));

		num_items = conf_section_numberof_items(section);

		for(j = 0; j < num_items; j++)
		{
			item = conf_section_get_item(section, j);

			fprintf(fp, "%s=%s\n", conf_section_item_get_name(item), conf_section_item_get_value(item));
		}
	}

	fflush(fp); sync();

	fclose(fp);

	return  0;

ERROR:
	return -1;
}

conf_section_t *
conf_append_section(conf_t *conf, const char *name)
{
	conf_section_t *section;

	if(!conf || strutil_isempty(name))
	{
		goto ERROR;
	}

	section = conf_section_create();

	conf_section_set_name(section, name);

	return list_add(conf->sections, section, list_size(conf->sections));

ERROR:
	return NULL;
}

void
conf_remove_section(conf_t *conf, int index)
{
	if(!conf)
	{
		goto ERROR;
	}

	conf_section_destroy(list_remove(conf->sections, index));

ERROR:
	return;
}

void
conf_remove_section_by_name(conf_t *conf, const char *name)
{
	int size, i;
	conf_section_t *section;

	if(!conf || strutil_isempty(name))
	{
		goto ERROR;
	}

	size = list_size(conf->sections);

	for(i = 0; i < size; i++)
	{
		section = list_get(conf->sections, i);

		if(strutil_isequal(section->name, name))
		{
			conf_section_destroy(list_remove(conf->sections, i));

			break;
		}
	}

ERROR:
	return;
}

void
conf_remove_sections_by_name(conf_t *conf, const char *name)
{
	int size, i;
	conf_section_t *section;

	if(!conf || strutil_isempty(name))
	{
		goto ERROR;
	}

	size = list_size(conf->sections);

	for(i = 0; i < size; i++)
	{
		section = list_get(conf->sections, i);

		if(strutil_isequal(section->name, name))
		{
			conf_section_destroy(list_remove(conf->sections, i));
		}
	}

ERROR:
	return;
}

conf_section_t *
conf_get_section(const conf_t *conf, int index)
{
	return conf ? list_get(conf->sections, index) : NULL;
}

conf_section_t *
conf_get_section_by_name(const conf_t *conf, const char *name)
{
	int size, i;
	conf_section_t *section;

	if(!conf || strutil_isempty(name))
	{
		goto ERROR;
	}

	size = list_size(conf->sections);

	for(i = 0; i < size; i++)
	{
		section = list_get(conf->sections, i);

		if(strutil_isequal(section->name, name))
		{
			return section;
		}
	}

ERROR:
	return NULL;
}

list_t *
conf_get_sections_by_name(const conf_t *conf, const char *name)
{
	list_t *dst;
	int size, i;
	conf_section_t *section;

	if(!conf || strutil_isempty(name))
	{
		goto ERROR;
	}

	dst = list_create();

	size = list_size(conf->sections);

	for(i = 0; i < size; i++)
	{
		section = list_get(conf->sections, i);

		if(strutil_isequal(section->name, name))
		{
			list_add(dst, section, list_size(dst));
		}
	}

	return dst;

ERROR:
	return NULL;
}

static conf_section_t *
conf_section_create(void)
{
	conf_section_t *section;

	section = (conf_section_t *)malloc(sizeof(conf_section_t));

	if(section)
	{
		section->name = NULL;

		section->items = list_create();
	}

	return section;
}

static void
conf_section_destroy(conf_section_t *section)
{
	if(section)
	{
		free(section->name);

		while(list_size(section->items))
		{
			conf_section_item_destroy(list_remove(section->items, 0));
		}

		list_destroy(section->items);
	}

	free(section);
}

int
conf_section_numberof_items(const conf_section_t *section)
{
	return section ? list_size(section->items) : -1;
}

char *
conf_section_get_name(const conf_section_t *section)
{
	return section ? section->name : NULL;
}

char *
conf_section_set_name(conf_section_t *section, const char *name)
{
	return (section && !strutil_isempty(name)) ? (section->name = strutil_dup(name)) : NULL;
}

conf_section_item_t *
conf_section_append_item(conf_section_t *section, const char *name, const char *value)
{
	conf_section_item_t *item;

	if(!section || strutil_isempty(name))
	{
		goto ERROR;
	}

	item = conf_section_item_create();

	conf_section_item_set_name(item, name);

	conf_section_item_set_value(item, value);

	return list_add(section->items, item, list_size(section->items));

ERROR:
	return NULL;
}

void
conf_section_remove_item(conf_section_t *section, int index)
{
	if(!section)
	{
		goto ERROR;
	}

	conf_section_item_destroy(list_remove(section->items, index));

ERROR:
	return;
}

void
conf_section_remove_item_by_name(conf_section_t *section, const char *name)
{
	int size, i;
	conf_section_item_t *item;

	if(!section || strutil_isempty(name))
	{
		goto ERROR;
	}

	size = list_size(section->items);

	for(i = 0; i < size; i++)
	{
		item = list_get(section->items, i);

		if(strutil_isequal(item->name, name))
		{
			conf_section_item_destroy(list_remove(section->items, i));

			break;
		}
	}

ERROR:
	return;
}

void
conf_section_remove_items_by_name(conf_section_t *section, const char *name)
{
	int size, i;
	conf_section_item_t *item;

	if(!section || strutil_isempty(name))
	{
		goto ERROR;
	}

	size = list_size(section->items);

	for(i = 0; i < size; i++)
	{
		item = list_get(section->items, i);

		if(strutil_isequal(item->name, name))
		{
			conf_section_item_destroy(list_remove(section->items, i));
		}
	}

ERROR:
	return;
}

conf_section_item_t *
conf_section_get_item(const conf_section_t *section, int index)
{
	return section ? list_get(section->items, index) : NULL;
}

conf_section_item_t *
conf_section_get_item_by_name(const conf_section_t *section, const char *name)
{
	int size, i;
	conf_section_item_t *item;

	if(!section || strutil_isempty(name))
	{
		goto ERROR;
	}

	size = list_size(section->items);

	for(i = 0; i < size; i++)
	{
		item = list_get(section->items, i);

		if(strutil_isequal(item->name, name))
		{
			return item;
		}
	}

ERROR:
	return NULL;
}

list_t *
conf_section_get_items_by_name(const conf_section_t *section, const char *name)
{
	list_t *dst;
	int size, i;
	conf_section_item_t *item;

	if(!section || strutil_isempty(name))
	{
		goto ERROR;
	}

	dst = list_create();

	size = list_size(section->items);

	for(i = 0; i < size; i++)
	{
		item = list_get(section->items, i);

		if(strutil_isequal(item->name, name))
		{
			list_add(dst, item, list_size(dst));
		}
	}

	return dst;

ERROR:
	return NULL;
}

static conf_section_item_t *
conf_section_item_create(void)
{
	conf_section_item_t *item;

	item = (conf_section_item_t *)malloc(sizeof(conf_section_item_t));

	if(item)
	{
		item->name = NULL;

		item->value = NULL;
	}

	return item;
}

static void
conf_section_item_destroy(conf_section_item_t *item)
{
	if(item)
	{
		free(item->name);

		free(item->value);
	}

	free(item);
}

char *
conf_section_item_get_name(const conf_section_item_t *item)
{
	return item ? item->name : NULL;
}

char *
conf_section_item_get_value(const conf_section_item_t *item)
{
	return item ? item->value : NULL;
}

char *
conf_section_item_set_name(conf_section_item_t *item, const char *name)
{
	return (item && !strutil_isempty(name)) ? strutil_exchange(&item->name, name) : NULL;
}

char *
conf_section_item_set_value(conf_section_item_t *item, const char *value)
{
	return item ? strutil_exchange(&item->value, value) : NULL;
}
