/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <poll.h>
#include <errno.h> 
#include <ctype.h>
#include "ftp.h"
#include "strutil.h"
#include "debug.h"

#define FTP_MODE_ACTIVE  0
#define FTP_MODE_PASSIVE 1

struct ftp
{
	int csoc;
	int dsoc;
	int pasv;
	char replymsg[1024];
	char *addr;
	char *port;
	char *user;
	char *pass;
	logger_t *logger;
};

static int
ftp_putcmd(ftp_t *ftp, const char *cmd, const char *arg);
static int
ftp_getreply(ftp_t *ftp);
static int
ftp_getline(int soc, char *dst, int size);
static int
ftp_auth(ftp_t *ftp);
static int
ftp_eprt(ftp_t *ftp);
static int
ftp_epsv(ftp_t *ftp);
static list_t *
ftp_nlst(ftp_t *ftp, const char *path);
static int
ftp_retr(ftp_t *ftp, const char *path, int fd);
static int
ftp_stor(ftp_t *ftp, const char *path, int fd);
static char *
_strndup(const char *src, int size);

#ifdef FTP_DEBUG
int
main(int argc, char *argv[])
{
	ftp_t *ftp;
	int err;

	ftp = ftp_create();

	if(!ftp)
	{
		goto ERROR;
	}

	ftp_setopt_addr(ftp, "192.168.1.12");
	ftp_setopt_port(ftp, "21");
	ftp_setopt_user(ftp, "root");
	ftp_setopt_pass(ftp, "rootroot");

	ftp->pasv = 1;

	err = ftp_open(ftp);

	if(err)
	{
		logger_printf(ftp->logger, LOGGER_ERROR, "ftp_open(): error");

		goto ERROR;
	}

	ftp_get(ftp, "/etc/passwd", "download.dat");

	ftp_put(ftp, "download.dat", "/tmp/hoge.txt");

	ftp_cd(ftp, "/tmp");

	free(ftp_pwd(ftp));

	ftp_close(ftp);

	ftp_destroy(ftp);

	return  0;

ERROR:
	ftp_destroy(ftp);

	return -1;
}
#endif

ftp_t *
ftp_create(void)
{
	ftp_t *ftp;

	ftp = (ftp_t *)malloc(sizeof(ftp_t));

	if(ftp)
	{
		ftp->csoc = -1;
		ftp->dsoc = -1;
		ftp->pasv =  1;
		ftp->addr = NULL;
		ftp->port = NULL;
		ftp->user = NULL;
		ftp->pass = NULL;
		ftp->logger = NULL;
	}

	return ftp;
}

void
ftp_destroy(ftp_t *ftp)
{
	if(ftp)
	{
		free(ftp->addr);
		free(ftp->port);
		free(ftp->user);
		free(ftp->pass);
	}

	free(ftp);
}

int
ftp_open(ftp_t *ftp)
{
	struct addrinfo hints, *res, *ptr;
	int err, soc = -1, ret;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;

	err = getaddrinfo(ftp->addr, ftp->port, &hints, &res);

	if(err)
	{
		logger_printf(ftp->logger, LOGGER_ERROR, "getaddrinfo(): %s", gai_strerror(err));

		goto ERROR;
	}

	for(ptr = res; ptr; ptr = ptr->ai_next)
	{
		soc = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol);

		if(soc < 0)
		{
			continue;
		}

		err = connect(soc, ptr->ai_addr, ptr->ai_addrlen);

		if(err)
		{
			close(soc);

			soc = -1;

			continue;
		}

		break;
	}

	freeaddrinfo(res);

	if(soc < 0)
	{
		goto ERROR;
	}

	ftp->csoc = soc;

	ret = ftp_getreply(ftp);

	if(ret != 220)
	{
		ftp_close(ftp);

		goto ERROR;
	}

	ret = ftp_auth(ftp);

	if(ret != 230)
	{
		ftp_close(ftp);

		goto ERROR;
	}

	ftp_putcmd(ftp, "TYPE", "I");

	ftp_getreply(ftp);

	return  0;

ERROR:
	return -1;
}

int
ftp_close(ftp_t *ftp)
{
	if(!ftp)
	{
		goto ERROR;
	}

	ftp_putcmd(ftp, "QUIT", NULL);

	ftp_getreply(ftp);

	return close(ftp->csoc);

ERROR:
	return -1;
}

static int
ftp_putcmd(ftp_t *ftp, const char *cmd, const char *arg)
{
	char buf[1024];
	int len, done = 0, ret;

	if(!ftp || !cmd)
	{
		goto ERROR;
	}

	if(arg)
	{
		sprintf(buf, "%s %s\r\n", cmd, arg);
	}
	else
	{
		sprintf(buf, "%s\r\n", cmd);
	}

	logger_printf(ftp->logger, LOGGER_DEBUG, ">>> %s", buf);

	len = strlen(buf);

	while(done < len)
	{
		ret = send(ftp->csoc, buf + done, len - done, 0);

		if(ret == -1)
		{
			if(errno == EINTR)
			{
				continue;
			}

			break;
		}

		if(ret == 0)
		{
			break;
		}

		done += ret;
	}

	return  0;

ERROR:
	return -1;
}

static int
ftp_getreply(ftp_t *ftp)
{
	struct pollfd pfd;
	char *ptr;
	int ready, len, code = -1;

	if(!ftp)
	{
		goto ERROR;
	}

	memset(ftp->replymsg, 0x00, sizeof(ftp->replymsg));

	pfd.fd = ftp->csoc;
	pfd.events = POLLIN;
	pfd.revents = 0;

	ptr = ftp->replymsg;

	while(1)
	{
		ready = poll(&pfd, 1, 10000);

		if(!ready)
		{
			if(errno == EINTR)
			{
				continue;
			}

			goto ERROR;
		}

		len = ftp_getline(ftp->csoc, ptr, sizeof(ftp->replymsg) - (ptr - ftp->replymsg));

		if(len == 0)
		{
			break;
		}

		if(len >= 6 && isdigit(*ptr) && isdigit(*(ptr + 1)) && isdigit(*(ptr + 2)) && *(ptr + 3) == ' ')
		{
			code = atoi(ptr);

			break;
		}

		ptr += len;
	}

	logger_printf(ftp->logger, LOGGER_DEBUG, "%s", ftp->replymsg);

	return code;

ERROR:
	return -1;
}

static int
ftp_getline(int soc, char *dst, int size)
{
	char *ptr;
	int ret;

	ptr = dst;

	while(ptr - dst < size)
	{
		ret = read(soc, ptr, 1);

		if(ret < 0)
		{
			if(errno == EINTR)
			{
				continue;
			}

			goto ERROR;
		}

		if(ret == 0)
		{
			break;
		}

		if(ptr != dst && *ptr == '\n' && *(ptr - 1) == '\r')
		{
			ptr++;

			break;
		}

		ptr++;
	}

	return ptr - dst;

ERROR:
	return -1;
}

static int
ftp_auth(ftp_t *ftp)
{
	int err, code;

	if(!ftp)
	{
		goto ERROR;
	}

	err = ftp_putcmd(ftp, "USER", ftp->user);

	if(err)
	{
		goto ERROR;
	}

	code = ftp_getreply(ftp);

	if(code == 331)
	{
		err = ftp_putcmd(ftp, "PASS", ftp->pass);

		if(err)
		{
			goto ERROR;
		}

		code = ftp_getreply(ftp);
	}

	return code;

ERROR:
	return -1;
}

static int
ftp_eprt(ftp_t *ftp)
{
	struct sockaddr_storage ss;
	socklen_t ss_len;;
	char host[NI_MAXHOST], serv[NI_MAXSERV], buf[1024];
	int err, soc, code;

	ss_len = sizeof(ss);

	memset(&ss, 0x00, ss_len);

	getsockname(ftp->csoc, (struct sockaddr *)&ss, &ss_len);

	err = getnameinfo((struct sockaddr *)&ss, ss_len, host, sizeof(host), NULL, 0, NI_NUMERICHOST);

	if(err)
	{
		logger_printf(ftp->logger, LOGGER_ERROR, "%s", gai_strerror(err));

		goto ERROR;
	}

	soc = socket(ss.ss_family, SOCK_STREAM, 0);

	if(soc < 0)
	{
		goto ERROR;
	}

	if(ss.ss_family == AF_INET)
	{
		((struct sockaddr_in *)&ss)->sin_port = 0;
	}
	else
	{
		((struct sockaddr_in6 *)&ss)->sin6_port = 0;
	}

	err = bind(soc, (struct sockaddr *)&ss, ss_len);

	if(err)
	{
		goto ERROR;
	}

	getsockname(soc, (struct sockaddr *)&ss, &ss_len);

	err = getnameinfo((struct sockaddr *)&ss, ss_len, host, sizeof(host), serv, sizeof(serv), NI_NUMERICHOST | NI_NUMERICSERV);

	if(err)
	{
		logger_printf(ftp->logger, LOGGER_ERROR, "%s", gai_strerror(err));

		goto ERROR;
	}

	sprintf(buf, "|%d|%s|%s|", ss.ss_family == AF_INET ? 1 : 2, host, serv);

	err = ftp_putcmd(ftp, "EPRT", buf);

	if(err)
	{
		goto ERROR;
	}

	code = ftp_getreply(ftp);

	if(code != 200)
	{
		struct sockaddr_in *sin;
		unsigned char *addr, *port;

		if(ss.ss_family != AF_INET)
		{
			goto ERROR;
		}

		sin = (struct sockaddr_in *)&ss;

		addr = (unsigned char *)&sin->sin_addr.s_addr;
		port = (unsigned char *)&sin->sin_port;

		sprintf(buf, "%u,%u,%u,%u,%u,%u", *addr, *(addr + 1), *(addr + 2), *(addr + 3), *port, *(port + 1));

		err = ftp_putcmd(ftp, "PORT", buf);

		if(err)
		{
			goto ERROR;
		}

		code = ftp_getreply(ftp);

		if(code != 200)
		{
			goto ERROR;
		}
	}

	listen(soc, 1);

	return soc;

ERROR:
	if(soc >= 0)
	{
		close(soc);
	}

	return -1;
}

static int
ftp_epsv(ftp_t *ftp)
{
	struct sockaddr_storage ss;
	socklen_t ss_len;;
	int err, code, ret, soc;

	if(!ftp)
	{
		goto ERROR;
	}

	ss_len = sizeof(ss);

	memset(&ss, 0x00, ss_len);

	getpeername(ftp->csoc, (struct sockaddr *)&ss, &ss_len);

	err = ftp_putcmd(ftp, "EPSV", NULL);

	if(err)
	{
		goto ERROR;
	}

	code = ftp_getreply(ftp);

	if(code != 229)
	{
		char addr1[4], addr2[4], addr3[4], addr4[4], port1[4], port2[4];

		if(ss.ss_family != AF_INET)
		{
			goto ERROR;
		}

		err = ftp_putcmd(ftp, "PASV", NULL);

		if(err)
		{
			goto ERROR;
		}

		code = ftp_getreply(ftp);

		if(code != 227)
		{
			goto ERROR;
		}

		ret = sscanf(ftp->replymsg, "%*[^(](%3[^,],%3[^,],%3[^,],%3[^,],%3[^,],%3[^)])\r\n", addr1, addr2, addr3, addr4, port1, port2);

		if(ret != 6)
		{
			goto ERROR;
		}

		((struct sockaddr_in *)&ss)->sin_addr.s_addr = htonl((atoi(addr1) << 24) + (atoi(addr2) << 16) + (atoi(addr3) << 8) + atoi(addr4));
		((struct sockaddr_in *)&ss)->sin_port = htons((atoi(port1) << 8) + atoi(port2));
	}
	else
	{
		char port[6];

		ret = sscanf(ftp->replymsg, "%*[^(](|||%5[^|]|)\r\n", port);

		if(ret != 1)
		{
			goto ERROR;
		}

		if(ss.ss_family == AF_INET)
		{
			((struct sockaddr_in *)&ss)->sin_port = htons(atoi(port));
		}
		else
		{
			((struct sockaddr_in6 *)&ss)->sin6_port = htons(atoi(port));
		}
	}

	soc = socket(ss.ss_family, SOCK_STREAM, 0);

	if(soc < 0)
	{
		goto ERROR;
	}

	err = connect(soc, (struct sockaddr *)&ss, ss_len);

	if(err)
	{
		goto ERROR;
	}

	return  soc;

ERROR:
	if(soc >= 0)
	{
		close(soc);
	}

	return -1;
}

static list_t *
ftp_nlst(ftp_t *ftp, const char *path)
{
	int soc = -1, acc = -1, code, ready, len;
	struct sockaddr_storage ss;
	socklen_t ss_len;
	struct pollfd pfd;
	char buf[1024];
	list_t *list;

	if(!ftp)
	{
		goto ERROR;
	}

	soc = ftp->pasv ? ftp_epsv(ftp) : ftp_eprt(ftp);

	if(soc < 0)
	{
		goto ERROR;
	}

	ftp_putcmd(ftp, "NLST", path);

	code = ftp_getreply(ftp);

	if(code != 150)
	{
		goto ERROR;
	}

	if(ftp->pasv)
	{
		acc = soc;

		soc = -1;
	}
	else
	{
		ss_len = sizeof(ss);
		memset(&ss, 0x00, ss_len);

		acc = accept(soc, (struct sockaddr *)&ss, &ss_len);

		if(acc < 0)
		{
			goto ERROR;
		}

		close(soc);

		soc = -1;
	}
		
	list = list_create();

	pfd.fd = acc;
	pfd.events = POLLIN;
	pfd.revents = 0;

	while(1)
	{
		ready = poll(&pfd, 1, 10000);

		if(!ready)
		{
			if(errno == EINTR)
			{
				continue;
			}

			goto ERROR;
		}

		len = ftp_getline(acc, buf, sizeof(buf));

		if(len == 0)
		{
			break;
		}

		buf[len] = '\0';

		list_add(list, _strndup(buf, strlen(buf) - 2), list_size(list));
	}

	code = ftp_getreply(ftp);

	close(acc);

	return list;

ERROR:
	if(soc >= 0)
	{
		close(soc);
	}

	if(acc >= 0)
	{
		close(acc);
	}

	if(list)
	{
		while(list_size(list))
		{
			free(list_remove(list, 0));
		}

		list_destroy(list);
	}

	return NULL;
}

static int
ftp_retr(ftp_t *ftp, const char *path, int fd)
{
	int soc = -1, acc = -1, code, ready, done = 0, len;
	struct sockaddr_storage ss;
	socklen_t ss_len;
	struct pollfd pfd;
	char buf[1024];

	if(!ftp)
	{
		goto ERROR;
	}

	soc = ftp->pasv ? ftp_epsv(ftp) : ftp_eprt(ftp);

	if(soc < 0)
	{
		goto ERROR;
	}

	ftp_putcmd(ftp, "RETR", path);

	code = ftp_getreply(ftp);

	if(code != 150)
	{
		goto ERROR;
	}

	if(ftp->pasv)
	{
		acc = soc;

		soc = -1;
	}
	else
	{
		ss_len = sizeof(ss);
		memset(&ss, 0x00, ss_len);

		acc = accept(soc, (struct sockaddr *)&ss, &ss_len);

		if(acc < 0)
		{
			goto ERROR;
		}

		close(soc);

		soc = -1;
	}
		
	pfd.fd = acc;
	pfd.events = POLLIN;
	pfd.revents = 0;

	while(1)
	{
		ready = poll(&pfd, 1, 10000);

		if(!ready)
		{
			if(errno == EINTR)
			{
				continue;
			}

			goto ERROR;
		}

		len = ftp_getline(acc, buf, sizeof(buf));

		if(len == 0)
		{
			break;
		}

		write(fd, buf, len);

		done += len;
	}

	ftp_getreply(ftp);

	close(acc);

	return done;

ERROR:
	if(soc >= 0)
	{
		close(soc);
	}

	if(acc >= 0)
	{
		close(acc);
	}

	return -1;
}

static int
ftp_stor(ftp_t *ftp, const char *path, int fd)
{
	int soc = -1, acc = -1, code, done = 0, len;
	struct sockaddr_storage ss;
	socklen_t ss_len;
	char buf[1024];

	if(!ftp)
	{
		goto ERROR;
	}

	soc = ftp->pasv ? ftp_epsv(ftp) : ftp_eprt(ftp);

	if(soc < 0)
	{
		goto ERROR;
	}

	ftp_putcmd(ftp, "STOR", path);

	code = ftp_getreply(ftp);

	if(code != 150)
	{
		goto ERROR;
	}

	if(ftp->pasv)
	{
		acc = soc;

		soc = -1;
	}
	else
	{
		ss_len = sizeof(ss);
		memset(&ss, 0x00, ss_len);

		acc = accept(soc, (struct sockaddr *)&ss, &ss_len);

		if(acc < 0)
		{
			goto ERROR;
		}

		close(soc);

		soc = -1;
	}
		
	while(1)
	{
		len = read(fd, buf, sizeof(buf));

		if(len < 0)
		{
			if(errno == EINTR)
			{
				continue;
			}

			goto ERROR;
		}

		if(len == 0)
		{
			break;
		}

		send(acc, buf, len, 0);

		done += len;
	}

	close(acc);

	ftp_getreply(ftp);

	return done;

ERROR:
	if(soc >= 0)
	{
		close(soc);
	}

	if(acc >= 0)
	{
		close(acc);
	}

	return -1;
}

char *
ftp_setopt_addr(ftp_t *ftp, const char *addr)
{
	if(!ftp)
	{
		goto ERROR;
	}

	free(ftp->addr);

	ftp->addr = addr ? strdup(addr) : NULL;

	return ftp->addr;

ERROR:
	return NULL;
}

char *
ftp_setopt_port(ftp_t *ftp, const char *port)
{
	if(!ftp)
	{
		goto ERROR;
	}

	free(ftp->port);

	ftp->port = port ? strdup(port) : NULL;

	return ftp->port;

ERROR:
	return NULL;
}

char *
ftp_setopt_user(ftp_t *ftp, const char *user)
{
	if(!ftp)
	{
		goto ERROR;
	}

	free(ftp->user);

	ftp->user = user ? strdup(user) : NULL;

	return ftp->user;

ERROR:
	return NULL;
}

char *
ftp_setopt_pass(ftp_t *ftp, const char *pass)
{
	if(!ftp)
	{
		goto ERROR;
	}

	free(ftp->pass);

	ftp->pass = pass ? strdup(pass) : NULL;

	return ftp->pass;

ERROR:
	return NULL;
}

logger_t *
ftp_setopt_logger(ftp_t *ftp, logger_t *logger)
{
	if(!ftp)
	{
		goto ERROR;
	}

	ftp->logger = logger;

	return ftp->logger;

ERROR:
	return NULL;
}

static char *
_strndup(const char *src, int size)
{
	char *dst;
	
	if(!src || size < 0)
	{       
		goto ERROR;
	}
	
	dst = (char *)calloc(size + 1, sizeof(char));
	
	if(!dst)
	{       
		goto ERROR;
	}
	
	return strncpy(dst, src, size);

ERROR:  
	return NULL;
}

int
ftp_cd(ftp_t *ftp, const char *path)
{
	int err, code;

	if(!ftp)
	{
		goto ERROR;
	}

	err = ftp_putcmd(ftp, "CWD", path);

	if(err)
	{
		goto ERROR;
	}

	code = ftp_getreply(ftp);

	if(code != 250)
	{
		goto ERROR;
	}

	return  0;

ERROR:
	return -1;
}

int
ftp_mkdir(ftp_t *ftp, const char *path)
{
	int err, code;

	if(!ftp)
	{
		goto ERROR;
	}

	err = ftp_putcmd(ftp, "MKD", path);

	if(err)
	{
		goto ERROR;
	}

	code = ftp_getreply(ftp);

	if(code != 257)
	{
		goto ERROR;
	}

	return  0;

ERROR:
	return -1;
}

char *
ftp_pwd(ftp_t *ftp)
{
	int err, code, ret;
	char buf[1024];

	if(!ftp)
	{
		goto ERROR;
	}

	err = ftp_putcmd(ftp, "PWD", NULL);

	if(err)
	{
		goto ERROR;
	}

	code = ftp_getreply(ftp);

	if(code != 257)
	{
		goto ERROR;
	}

	ret = sscanf(ftp->replymsg, "%*[^\"]\"%1023[^\"]\"\r\n", buf);

	if(ret != 1)
	{
		goto ERROR;
	}

	return strdup(buf);

ERROR:
	return NULL;
}

list_t *
ftp_nlist(ftp_t *ftp, const char *path)
{
	return ftp_nlst(ftp, path);
}

int
ftp_get(ftp_t *ftp, const char *remote, const char *local)
{
	int fd, len;

	fd = open(local, O_WRONLY | O_CREAT | O_TRUNC, 0644);

	if(fd < 0)
	{
		logger_printf(ftp->logger, LOGGER_DEBUG, "    can't open local file '%s'", local);

		goto ERROR;
	}

	len = ftp_retr(ftp, remote, fd);

	close(fd);

	return len;

ERROR:
	return -1;
}

int
ftp_put(ftp_t *ftp, const char *local, const char *remote)
{
	int fd, len, err;
	char *pwd, *ptr, *dir;

	if(!ftp || !remote)
	{
		goto ERROR;
	}

	pwd = ftp_pwd(ftp);

	ptr = (char *)remote;

	while(1)
	{
		ptr = strstr(ptr, "/");

		if(!ptr)
		{
			break;
		}

		ptr++;

		dir = strutil_ndup(remote, ptr - remote);

		err = ftp_cd(ftp, dir);

		if(err)
		{
			err = ftp_mkdir(ftp, dir);

			if(err)
			{
				free(dir);

				goto ERROR;
			}

			err = ftp_cd(ftp, dir);

			if(err)
			{
				free(dir);

				goto ERROR;
			}
		}

		free(dir);
	}

	fd = open(local, O_RDONLY);

	if(fd < 0)
	{
		logger_printf(ftp->logger, LOGGER_DEBUG, "    can't open local file '%s'", local);

		goto ERROR;
	}

	ftp_cd(ftp, pwd);

	len = ftp_stor(ftp, remote, fd);

	close(fd);

	return len;

ERROR:
	if(pwd)
	{
		ftp_cd(ftp, pwd);
	}

	return -1;
}
