/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdio.h>
#include <stdlib.h>
#include "list.h"
#include "debug.h"

typedef struct list_node list_node_t;

struct list
{
	list_node_t *head;
	list_node_t *tail;
	int size;
};

struct list_node
{
	void *data;
	list_node_t *next;
	list_node_t *prev;
};

#ifdef LIST_UNIT_TEST
int
main(int argc, char *argv[])
{
	list_t *list;
	int len, i;

	list = list_create();

	if(!list)
	{
		perror("list_create()");

		goto ERROR;
	}

	list_add(list, "hoge0", list_size(list));
	list_add(list, "hoge1", list_size(list));
	list_add(list, "hoge2", list_size(list));
	list_add(list, "hoge3", list_size(list));
	list_add(list, "hoge4", list_size(list));
	list_add(list, "hoge5", list_size(list));
	list_add(list, "hoge6", list_size(list));
	list_add(list, "hoge7", list_size(list));
	list_add(list, "hoge8", list_size(list));
	list_add(list, "hoge9", list_size(list));

	list_remove(list, 0);
	list_remove(list, 3);
	list_remove(list, list_size(list) - 1);

	list_swap(list, 0, list_size(list) - 1);
	list_swap(list, 0, 1);
	list_swap(list, list_size(list) - 1, 1);

	list_move(list, list_size(list) - 1, list_size(list) - 2);

	len = list_size(list);

	printf("list_size():%d\n", len);

	while(list_size(list))
	{
		printf("list = %s\n", (char *)list_remove(list, 0));
	}

	list_destroy(list);

	return  0;

ERROR:
	return -1;
}
#endif

list_t *
list_create(void)
{
	list_t *list;

	list = (list_t *)malloc(sizeof(list_t));

	if(list)
	{
		list->head = NULL;
		list->tail = NULL;
		list->size = 0;
	}

	return list;
}

void
list_destroy(list_t *list)
{
	if(list)
	{
		while(list_size(list))
		{
			list_remove(list, 0);
		}
	}

	free(list);
}

int
list_size(list_t *list)
{
	return list ? list->size : -1;
}

void *
list_add(list_t *list, void *data, int index)
{
	list_node_t *ptr, *node;
	int i;

	if(!list || list->size < index)
	{
		goto ERROR;
	}

	node = (list_node_t *)malloc(sizeof(list_t));

	if(!node)
	{
		goto ERROR;
	}

	node->data = data;

	ptr = list->head;

	for(i = 0; i < index; i++)
	{
		ptr = ptr->next;
	}

	node->next = ptr ? ptr : (index ? NULL : list->head);
	node->prev = ptr ? ptr->prev : (index ? list->tail : NULL);

	if(node->next)
	{
		node->next->prev = node;
	}
	else
	{
		list->tail = node;
	}

	if(node->prev)
	{
		node->prev->next = node;
	}
	else
	{
		list->head = node;
	}

	list->size += 1;

	return node->data;

ERROR:
	return NULL;
}

void *
list_remove(list_t *list, int index)
{
	list_node_t *ptr;
	int i;
	void *dst;

	if(!list || list->size <= index)
	{
		goto ERROR;
	}

	ptr = list->head;

	for(i = 0; i < index; i++)
	{
		ptr = ptr->next;
	}

	if(ptr->prev)
	{
		ptr->prev->next = ptr->next;
	}
	else
	{
		list->head = ptr->next;
	}

	if(ptr->next)
	{
		ptr->next->prev = ptr->prev;
	}
	else
	{
		list->tail = ptr->prev;
	}

	dst = ptr->data;

	list->size -= 1;

	free(ptr);

	return dst;

ERROR:
	return NULL;
}

void *
list_get(list_t *list, int index)
{
	list_node_t *ptr;
	int i;

	if(!list || list->size <= index)
	{
		goto ERROR;
	}

	ptr = list->head;

	for(i = 0; i < index; i++)
	{
		ptr = ptr->next;
	}

	return ptr->data;

ERROR:
	return NULL;
}

void
list_move(list_t *list, int src, int dst)
{
	list_node_t *sptr, *dptr;
	int i;

	if(!list || list->size <= src || src < 0  || list->size <= dst || dst < 0)
	{
		goto ERROR;
	}

	sptr = list->head;

	for(i = 0; i < src; i++)
	{
		sptr = sptr->next;
	}

	dptr = list->head;

	for(i = 0; i < dst; i++)
	{
		dptr = dptr->next;
	}

	if(sptr->next)
	{
		sptr->next->prev = sptr->prev;
	}

	if(sptr->prev)
	{
		sptr->prev->next = sptr->next;
	}

	if(dptr->prev)
	{
		dptr->prev->next = sptr;
	}

	sptr->next = dptr;

	sptr->prev = dptr->prev;

	dptr->prev = sptr;

	if(!sptr->prev)
	{
		list->head = sptr;
	}

	if(!sptr->next)
	{
		list->tail = sptr;
	}

	if(!dptr->next)
	{
		list->tail = dptr;
	}

	return;

ERROR:
	return;
}

void
list_swap(list_t *list, int n1, int n2)
{
	list_node_t *ptr1, *ptr2, *tmp;
	int i;

	if(!list || list->size <= n1 || list->size <= n2)
	{
		goto ERROR;
	}

	ptr1 = list->head;

	for(i = 0; i < n1; i++)
	{
		ptr1 = ptr1->next;
	}

	ptr2 = list->head;

	for(i = 0; i < n2; i++)
	{
		ptr2 = ptr2->next;
	}

	if(ptr1->next && ptr1->next != ptr2)
	{
		ptr1->next->prev = ptr2;
	}

	if(ptr1->prev && ptr1->prev != ptr2)
	{
		ptr1->prev->next = ptr2;
	}

	if(ptr2->next && ptr2->next != ptr1)
	{
		ptr2->next->prev = ptr1;
	}

	if(ptr2->prev && ptr2->prev != ptr1)
	{
		ptr2->prev->next = ptr1;
	}

	tmp = ptr1->next;

	ptr1->next = (ptr2->next != ptr1) ? ptr2->next : ptr2;
	ptr2->next = (tmp != ptr2) ? tmp : ptr1;

	tmp = ptr1->prev;

	ptr1->prev = (ptr2->prev != ptr1) ? ptr2->prev : ptr2;
	ptr2->prev = (tmp != ptr2) ? tmp : ptr1;

	if(!ptr1->prev)
	{
		list->head = ptr1;
	}

	if(!ptr1->next)
	{
		list->tail = ptr1;
	}

	if(!ptr2->prev)
	{
		list->head = ptr2;
	}

	if(!ptr2->next)
	{
		list->tail = ptr2;
	}

	return;

ERROR:
	return;
}
