/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <ctype.h>
#include "logger.h"
#include "debug.h"

struct logger
{
	char *name;
	char *file;
	unsigned int size;
	int indent;
};

static void
logger_hexdump_core(FILE *fp, const char *head, const char *data, int len, int indent);
static int
logger_printf_core(FILE *fp, const char *head, const char *body, int indent);
static void
logger_print_indent(FILE *fp, int indent);

logger_t *glogger = NULL;

#ifdef _LOGGER_DEBUG
int
main(int argc, char *argv[])
{
	logger_t *logger;

	logger = logger_create();

	logger->name = strdup("logger");
	logger->file = strdup("./log");

	logger_printf(logger, LOGGER_DEBUG, "logger test 1");
	logger_printf(logger, LOGGER_DEBUG, "logger test 2");
	logger_printf(logger, LOGGER_DEBUG, "logger test 3");
	logger_printf(logger, LOGGER_DEBUG, "logger test 4-1\nlogger test 4-2");
	logger_printf(logger, LOGGER_DEBUG, "logger test 5\n");
	logger_printf(logger, LOGGER_DEBUG, "logger test 6\nlogger test 7\n");
	logger_printf(logger, LOGGER_DEBUG, "logger test 8");
	logger_printf(logger, LOGGER_DEBUG, "\nlogger test 9");
	logger_printf(logger, LOGGER_DEBUG, "");
	logger_printf(logger, LOGGER_DEBUG, "");
	logger_printf(logger, LOGGER_DEBUG, "logger test 10");

	char buf[] = "1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ\n\r\t";

	logger_hexdump(logger, LOGGER_DEBUG, buf, sizeof(buf));

	logger_destroy(logger);

	return 0;
}
#endif

logger_t *
logger_create(void)
{
	logger_t *logger;

	logger = (logger_t *)malloc(sizeof(logger_t));

	if(logger)
	{
		logger->name = NULL;
		logger->file = NULL;
		logger->size = 1024 * 1024;
		logger->indent = 0;
	}

	return logger;
}

void
logger_destroy(logger_t *logger)
{
	if(logger)
	{
		free(logger->name);
		free(logger->file);
	}

	free(logger);
}

int
logger_open(logger_t *logger)
{
	return 0;
}

void
logger_close(logger_t *logger)
{
	//
}

char *
logger_setopt_name(logger_t *logger, const char *name)
{
	if(!logger)
	{
		goto ERROR;
	}

	if(logger->name)
	{
		free(logger->name);
	}

	logger->name = name ? strdup(name) : NULL;

	return logger->name;

ERROR:
	return NULL;
}

char *
logger_setopt_file(logger_t *logger, const char *file)
{
	if(!logger)
	{
		goto ERROR;
	}

	if(logger->file)
	{
		free(logger->file);
	}

	logger->file = file ? strdup(file) : NULL;

	return logger->file;

ERROR:
	return NULL;
}

void
_logger_entry_function(logger_t *logger, const char *function)
{
	logger_printf(logger, LOGGER_TRACE, "entry function: %s", function);

	logger->indent++;
}

void
_logger_leave_function(logger_t *logger, const char *function)
{
	logger->indent--;

	logger_printf(logger, LOGGER_TRACE, "leave function: %s", function);
}

int
logger_printf(logger_t *logger, int level, const char *format,...)
{
	char lv, now[128], head[1024], buf[4096];
	struct timeval tv;
	FILE *fp;
	va_list arg;
	int len;

	if(!logger || !logger->file)
	{
		goto ERROR;
	}

	switch(level)
	{
		case LOGGER_ERROR:
			lv = 'E'; break;
		case LOGGER_WARNING:
			lv = 'W'; break;
		case LOGGER_INFO:
			lv = 'I'; break;
		case LOGGER_DEBUG:
			lv = 'D'; break;
		case LOGGER_TRACE:
			lv = 'T'; break;
		default:
			goto ERROR;
	}

	gettimeofday(&tv, NULL);

	strftime(now, sizeof(now), "%Y/%m/%d %H:%M:%S", localtime(&tv.tv_sec));

	fp = fopen(logger->file, "ab");

	if(!fp)
	{
		goto ERROR;
	}

	snprintf(head, sizeof(head), "%s.%03d %s (%d) [%c] : ", now, (int)(tv.tv_usec / 1000), logger->name, getpid(), lv);

	fwrite(head, sizeof(char), strlen(head), fp);

	va_start(arg, format);

	len = vsnprintf(buf, sizeof(buf), format, arg);

	va_end(arg);

	if(len < 0)
	{
		goto ERROR;
	}

	logger_printf_core(fp, head, buf, logger->indent);

	fclose(fp);

	return  0;

ERROR:
	return -1;
}

void
logger_hexdump(logger_t *logger, int level, const void *data, int size)
{
	char lv, now[128], head[1024];
	struct timeval tv;
	FILE *fp;

	if(!logger || !logger->file)
	{
		goto ERROR;
	}

	switch(level)
	{
		case LOGGER_ERROR:
			lv = 'E'; break;
		case LOGGER_WARNING:
			lv = 'W'; break;
		case LOGGER_INFO:
			lv = 'I'; break;
		case LOGGER_DEBUG:
			lv = 'D'; break;
		case LOGGER_TRACE:
			lv = 'T'; break;
		default:
			goto ERROR;
	}

	gettimeofday(&tv, NULL);

	strftime(now, sizeof(now), "%Y/%m/%d %H:%M:%S", localtime(&tv.tv_sec));

	fp = fopen(logger->file, "ab");

	if(!fp)
	{
		goto ERROR;
	}

	snprintf(head, sizeof(head), "%s.%03d %s (%d) [%c] : ", now, (int)(tv.tv_usec / 1000), logger->name, getpid(), lv);

	logger_hexdump_core(fp, head, data, size, logger->indent);

	fclose(fp);

	return;

ERROR:
	return;
}

static void
logger_hexdump_core(FILE *fp, const char *head, const char *data, int len, int indent)
{
int  a, b;
char *ldat;

	ldat = (char *)calloc(len + 1, sizeof(char));

	for(a = 0; a < len; a++)
	{
		if(!isprint((unsigned char)(data[a])))
		{
			ldat[a] = '.';
		}
		else
		{
			ldat[a] = data[a];
		}
	}

	logger_print_indent(fp, indent);

	fprintf(fp, "%s+----------------------------------------------------------------------------\n", head);

	logger_print_indent(fp, indent);

	fprintf(fp, "%s| hexdump(): %dbyte\n", head, len);

	logger_print_indent(fp, indent);

	fprintf(fp, "%s+------+-------------------------------------------------+------------------+\n", head);

	for(a = 0; a < len; a = a + 16)
	{
		logger_print_indent(fp, indent);

		fprintf(fp, "%s| %04x | ", head, a);

		for(b = 0; b < 16; b++)
		{
			if(a + b < len)
			{
				fprintf(fp, "%02x ", 0xFF & data[a + b]);
			}
			else
			{
				fprintf(fp, "   ");
			}
		}

		fprintf(fp, "| ");

		for(b = 0; b < 16; b++)
		{
			if(a + b < len)
			{
				if(isascii((unsigned char)(ldat[a + b])) && isprint((unsigned char)(ldat[a + b])))
				{
					fprintf(fp, "%c", ldat[a + b]);
				}
				else
				{
					fprintf(fp, ".");
				}
			}
			else
			{
				fprintf(fp, " ");
			}
		}

		fprintf(fp, " |\n");
	}

	logger_print_indent(fp, indent);

	fprintf(fp, "%s+------+-------------------------------------------------+------------------+\n", head);

	free(ldat);
}

static int
logger_printf_core(FILE *fp, const char *head, const char *body, int indent)
{
	char *sp, *ep;

	if(!body)
	{
		goto ERROR;
	}

	sp = (char *)body;

	while(1)
	{
		ep = strstr(sp, "\n");

		if(!ep)
		{
			logger_print_indent(fp, indent);

			fwrite(sp, sizeof(char), strlen(sp), fp);

			break;
		}

		ep++;

		logger_print_indent(fp, indent);

		fwrite(sp, sizeof(char), ep - sp, fp);

		fwrite(head, sizeof(char), strlen(head), fp);

		sp = ep;
	}

	fwrite("\n", sizeof(char), 1, fp);

	return  0;

ERROR:
	return -1;
}

static void
logger_print_indent(FILE *fp, int indent)
{
	int i;
	char *tmp = "  ";

	for(i = indent; i > 0; i--)
	{
		fwrite(tmp, sizeof(char), strlen(tmp), fp);
	}
}
