/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pairarray.h"
#include "array.h"
#include "strutil.h"
#include "debug.h"

struct pair
{
	char *name;
	char *value;
};

struct pairarray
{
	array_t *array;
};

typedef struct pair pair_t;

static pair_t *
pair_create(void);
static void
pairarray_destroy_callback(void *obj);
static void *
pairarray_add_callback(const void *data);
static pair_t *
pairarray_get(const pairarray_t *array, int offset);

static pair_t *
pair_create(void)
{
pair_t *dst;

	dst = (pair_t *)malloc(sizeof(pair_t));

	if(dst)
	{
		dst->name  = NULL;
		dst->value = NULL;
	}

	return dst;
}

pairarray_t *
pairarray_create(void)
{
	return (pairarray_t *)array_create();
}

void
pairarray_destroy(pairarray_t *obj)
{
	return array_destroy((array_t *)obj, pairarray_destroy_callback);
}

static void
pairarray_destroy_callback(void *obj)
{
pair_t *ptr = obj;

	if(ptr)
	{
		free(ptr->name);

		free(ptr->value);
	}

	free(ptr);
}

char *
pairarray_add(pairarray_t *obj, const char *name, const char *value)
{
pair_t *tmp = NULL, *ret;

	if(!obj || !name || !strlen(name))
	{
		goto ERROR;
	}

	tmp = pair_create();

	if(!tmp)
	{
		goto ERROR;
	}

	tmp->name  = strdup(name);
	tmp->value = value ? strdup(value) : NULL;

	ret = array_add((array_t *)obj, (const void *)tmp, pairarray_add_callback);

	if(!ret)
	{
		goto ERROR;
	}

	return ret->value;

ERROR:
	pairarray_destroy_callback(tmp);

	return NULL;
}

static void *
pairarray_add_callback(const void *data)
{
	return (void *)data;
}

static pair_t *
pairarray_get(const pairarray_t *obj, int offset)
{
	return (pair_t *)array_get((array_t *)obj, offset);
}

char *
pairarray_get_name(const pairarray_t *obj, int offset)
{
pair_t *tmp;

        tmp = pairarray_get(obj, offset);

	if(!tmp)
	{
		goto ERROR;
	}

	return tmp->name;

ERROR:
	return NULL;
}

char *
pairarray_get_value(const pairarray_t *obj, int offset)
{
pair_t *tmp;

        tmp = pairarray_get(obj, offset);

	if(!tmp)
	{
		goto ERROR;
	}

	return tmp->value;

ERROR:
	return NULL;
}

char *
pairarray_get_value_by_name(const pairarray_t *obj, const char *name)
{
int     len, i;

	if(!name)
	{
		goto ERROR;
	}

	len = pairarray_len(obj);

	for(i = 0; i < len; i++)
	{
		if(strcmp(pairarray_get_name(obj, i), name) == 0)
		{
			return pairarray_get_value(obj, i);
		}
	}

ERROR:
	return NULL;
}

char *
pairarray_set_value(pairarray_t *obj, const char *value, int offset)
{
pair_t *tmp;

        tmp = pairarray_get(obj, offset);

	if(!tmp)
	{
		goto ERROR;
	}

	return strutil_exchange(&(tmp->value), value);

ERROR:
	return NULL;
}

char *
pairarray_set_value_by_name(pairarray_t *obj, const char *name, const char *value)
{
int len, i;

	if(!name)
	{
		goto ERROR;
	}

	len = pairarray_len(obj);

	for(i = 0; i < len; i++)
	{
		if(strcmp(pairarray_get_name(obj, i), name) == 0)
		{
			return pairarray_set_value(obj, value, i);
		}
	}

	return pairarray_add(obj, name, value);

ERROR:
	return NULL;
}

int
pairarray_len(const pairarray_t *obj)
{
        return array_len((array_t *)obj);
}

pairarray_t *
pairarray_dup(const pairarray_t *obj)
{
pairarray_t *dst;
int          len, i;

	if(!obj)
	{
		goto ERROR;
	}

	dst = pairarray_create();

	if(!dst)
	{
		goto ERROR;
	}

	len = pairarray_len(obj);

	for(i = 0; i < len; i++)
	{
		pairarray_add(dst, pairarray_get_name(obj, i), pairarray_get_value(obj, i));
	}

	return dst;

ERROR:
	return NULL;
}
