/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "property.h"
#include "strutil.h"
#include "strarray.h"
#include "debug.h"

struct property
{
	list_t *items;
};

struct property_item
{
	char *name;
	char *value;
};

static property_item_t *
property_item_create(void);
static void
property_item_destroy(property_item_t *item);

property_t *
property_create(void)
{
	property_t *property;

	property = (property_t *)malloc(sizeof(property_t));

	if(property)
	{
		property->items = list_create();
	}

	return property;
}

void
property_destroy(property_t *property)
{
	if(property)
	{
		while(list_size(property->items))
		{
			property_item_destroy(list_remove(property->items, 0));
		}

		list_destroy(property->items);
	}

	free(property);
}

int
property_numberof_items(const property_t *property)
{
	return property ? list_size(property->items) : -1;
}

int
property_load(property_t *property, const char *file)
{
	FILE *fp;
	char buf[4096], *name, *value;
	strarray_t *array;

	fp = fopen(file, "r");

	if(!fp)
	{
		goto ERROR;
	}

	while(fgets(buf, sizeof(buf), fp))
	{
		strutil_trim_direct(buf);

		if(strutil_isempty(buf) || buf[0] == '#')
		{
			continue;
		}

		array = strutil_split(buf, "=", 1);

		if(strarray_len(array) == 2)
		{
			name  = strutil_trim_direct(strarray_get(array, 0));
			value = strutil_trim_direct(strarray_get(array, 1));

			property_append_item(property, name, value);
		}

		strarray_destroy(array);
	}

	fclose(fp);

	return  0;

ERROR:
	return -1;
}

int
property_save(const property_t *property, const char *file)
{
	FILE *fp;
	int num, i;
	property_item_t *item;

	fp = fopen(file, "w");

	if(!fp)
	{
		goto ERROR;
	}

	num = property_numberof_items(property);

	for(i = 0; i < num; i++)
	{
		item = property_get_item(property, i);

		fprintf(fp, "%s=%s\n", property_item_get_name(item), property_item_get_value(item));
	}

	fflush(fp); sync();

	fclose(fp);

	return  0;

ERROR:
	return -1;
}

property_item_t *
property_append_item(property_t *property, const char *name, const char *value)
{
	property_item_t *item;

	if(!property || strutil_isempty(name))
	{
		goto ERROR;
	}

	item = property_item_create();

	property_item_set_name(item, name);

	property_item_set_value(item, value);

	return list_add(property->items, item, list_size(property->items));

ERROR:
	return NULL;
}

void
property_remove_item(property_t *property, int index)
{
	if(!property)
	{
		goto ERROR;
	}

	property_item_destroy(list_remove(property->items, index));

ERROR:
	return;
}

void
property_remove_item_by_name(property_t *property, const char *name)
{
	int size, i;
	property_item_t *item;

	if(!property || strutil_isempty(name))
	{
		goto ERROR;
	}

	size = list_size(property->items);

	for(i = 0; i < size; i++)
	{
		item = list_get(property->items, i);

		if(strutil_isequal(item->name, name))
		{
			property_item_destroy(list_remove(property->items, i));

			break;
		}
	}

ERROR:
	return;
}

void
property_remove_items_by_name(property_t *property, const char *name)
{
	int size, i;
	property_item_t *item;

	if(!property || strutil_isempty(name))
	{
		goto ERROR;
	}

	size = list_size(property->items);

	for(i = 0; i < size; i++)
	{
		item = list_get(property->items, i);

		if(strutil_isequal(item->name, name))
		{
			property_item_destroy(list_remove(property->items, i));
		}
	}

ERROR:
	return;
}

property_item_t *
property_get_item(const property_t *property, int index)
{
	return property ? list_get(property->items, index) : NULL;
}

property_item_t *
property_get_item_by_name(const property_t *property, const char *name)
{
	int size, i;
	property_item_t *item;

	if(!property || strutil_isempty(name))
	{
		goto ERROR;
	}

	size = list_size(property->items);

	for(i = 0; i < size; i++)
	{
		item = list_get(property->items, i);

		if(strutil_isequal(item->name, name))
		{
			return item;
		}
	}

ERROR:
	return NULL;
}

list_t *
property_get_items_by_name(const property_t *property, const char *name)
{
	list_t *dst;
	int size, i;
	property_item_t *item;

	if(!property || strutil_isempty(name))
	{
		goto ERROR;
	}

	dst = list_create();

	size = list_size(property->items);

	for(i = 0; i < size; i++)
	{
		item = list_get(property->items, i);

		if(strutil_isequal(item->name, name))
		{
			list_add(dst, item, list_size(dst));
		}
	}

	return dst;

ERROR:
	return NULL;
}

static property_item_t *
property_item_create(void)
{
	property_item_t *item;

	item = (property_item_t *)malloc(sizeof(property_item_t));

	if(item)
	{
		item->name = NULL;

		item->value = NULL;
	}

	return item;
}

static void
property_item_destroy(property_item_t *item)
{
	if(item)
	{
		free(item->name);

		free(item->value);
	}

	free(item);
}

char *
property_item_get_name(const property_item_t *item)
{
	return item ? item->name : NULL;
}

char *
property_item_get_value(const property_item_t *item)
{
	return item ? item->value : NULL;
}

char *
property_item_set_name(property_item_t *item, const char *name)
{
	return (item && !strutil_isempty(name)) ? strutil_exchange(&item->name, name) : NULL;
}

char *
property_item_set_value(property_item_t *item, const char *value)
{
	return item ? strutil_exchange(&item->value, value) : NULL;
}
