/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include "semlock.h"
#include "debug.h"

#define SEMLOCK_LOCK   -1
#define SEMLOCK_UNLOCK  1

struct semlock
{
	int semid;
};

#ifdef LINUX
union semun
{
	int              val;
	struct semid_ds *buf;
	unsigned short  *array;
	struct seminfo  *__buf;
};
#endif

static semlock_t *
semlock_create(void);
static void
semlock_destroy(semlock_t *obj);
static int
semlock_init(const semlock_t *obj);
static int
semlock_lock(const semlock_t *obj, int op);

static semlock_t *
semlock_create(void)
{
semlock_t *dst;

	dst = (semlock_t *)malloc(sizeof(semlock_t));

	if(dst)
	{
		dst->semid = -1;
	}

	return dst;
}

static void
semlock_destroy(semlock_t *obj)
{
	free(obj);
}

semlock_t *
semlock_open(key_t key)
{
semlock_t *dst;
int        ret;

	dst = semlock_create();

	if(!dst)
	{
		goto ERROR;
	}

	dst->semid = semget(key, 1, IPC_CREAT | IPC_EXCL | 0666);

	if(dst->semid == -1)
	{
		dst->semid = semget((key_t)key, 0, 0);

		if(dst->semid == -1)
		{
			goto ERROR;
		}
	}
	else
	{
		ret = semlock_init(dst);

		if(ret == -1)
		{
			goto ERROR;
		}
	}

	semlock_lock(dst, SEMLOCK_LOCK);

	return dst;

ERROR:
	semlock_destroy(dst);

	return NULL;
}

void
semlock_close(semlock_t *obj)
{
int ret;

	if(!obj)
	{
		return;
	}

	ret = semctl(obj->semid, 0, GETNCNT);

	switch(ret)
	{
		case  0:
			semctl(obj->semid, 0, IPC_RMID);

			break;

		default:
			semlock_lock(obj, SEMLOCK_UNLOCK);

			break;
	}

	semlock_destroy(obj);
}

static int
semlock_init(const semlock_t *obj)
{
union semun  semun;
int          ret;

	if(!obj)
	{
		goto ERROR;
	}

	ret = semctl(obj->semid, 0, GETVAL, semun);

	if(ret == -1)
	{
		goto ERROR;
	}

	semun.val = 1;

	ret = semctl(obj->semid, 0, SETVAL, semun);

	if(ret == -1)
	{
		goto ERROR;
	}

	return  0;

ERROR:
	return -1;
}

static int
semlock_lock(const semlock_t *obj, int op)
{
struct sembuf sembuf;

	if(!obj)
	{
		goto ERROR;
	}

	sembuf.sem_num = 0;
	sembuf.sem_op  = op;
	sembuf.sem_flg = SEM_UNDO;

	return semop(obj->semid, &sembuf, 1);

ERROR:
	return -1;
}
