/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "strutil.h"
#include "debug.h"

char *
strutil_tolower(char *str)
{
	char *ptr;

	if(str)
	{
		for(ptr = str; *ptr; ptr++)
		{
			*ptr = tolower(*ptr);
		}
	}

	return str;
}

char *
strutil_toupper(char *str)
{
	char *ptr;

	if(str)
	{
		for(ptr = str; *ptr; ptr++)
		{
			*ptr = toupper(*ptr);
		}
	}

	return str;
}

char *
strutil_dup(const char *src)
{
	return src ? strdup(src) : NULL;
}

char *
strutil_ndup(const char *src, int size)
{
	char *dst;

	if(!src || size < 0)
	{
		goto ERROR;
	}

	dst = (char *)calloc(size + 1, sizeof(char));

	if(!dst)
	{
		goto ERROR;
	}

	return strncpy(dst, src, size);

ERROR:
	return NULL;
}

char *
strutil_chomp(char *str)
{
	char *ptr;

	if(str && *str)
	{
		ptr = str + strlen(str) - 1;

		if(*ptr == '\n')
		{
			*ptr = '\0';

			if(ptr - str > 0)
			{
				ptr--;
			}
		}

		if(*ptr == '\r')
		{
			*ptr = '\0';
		}
	}

	return str;
}

char *
strutil_trim(const char *src)
{
	char *sp, *ep;

	if(!src)
	{
		goto ERROR;
	}

	for(sp = (char *)src; *sp; sp++)
	{
		if(!isspace(*sp))
		{
			break;
		}
	}

	for(ep = (char *)(src + (strlen(src))); ep > sp; ep--)
	{
		if(!isspace(*(ep - 1)))
		{
			break;
		}
	}

	return strutil_ndup(sp, ep - sp);

ERROR:
	return NULL;
}

char *
strutil_trim_direct(char *str)
{
	char *sp, *ep;

	if(!str)
	{
		goto ERROR;
	}

	for(sp = (char *)str; *sp; sp++)
	{
		if(!isspace(*sp))
		{
			break;
		}
	}

	for(ep = (char *)(str + (strlen(str))); ep > sp; ep--)
	{
		if(!isspace(*(ep - 1)))
		{
			break;
		}
	}

	memmove(str, sp, ep - sp);

	str[ep - sp] = '\0';

	return str;

ERROR:
	return NULL;
}

char *
strutil_exchange(char **dst, const char *src)
{
	if(!dst)
	{
		goto ERROR;
	}

	free(*dst);

	*dst = strutil_dup(src);

	return *dst;

ERROR:
	return NULL;
}

char *
strutil_unit(const char *str1, const char *str2, const char *glue)
{
	char *dst, *tmp1, *tmp2;

	if(!str1 && !str2 && !glue)
	{
		goto ERROR;
	}

	tmp1 = str1 ? (char *)str1 : "";
	tmp2 = str2 ? (char *)str2 : "";

	if(glue)
	{
		dst = (char *)calloc(strlen(tmp1) + strlen(glue) + strlen(tmp2) + 1, sizeof(char));

		if(!dst)
		{
			goto ERROR;
		}

		sprintf(dst, "%s%s%s", tmp1, glue, tmp2);
	}
	else
	{
		dst = (char *)calloc(strlen(tmp1) + strlen(tmp2) + 1, sizeof(char));

		if(!dst)
		{
			goto ERROR;
		}

		sprintf(dst, "%s%s", tmp1, tmp2);
	}

	return dst;

ERROR:
	return NULL;
}

strarray_t *
strutil_split(const char *src, const char *delim, int num)
{
	strarray_t *dst;
	char *sp, *ep, *tmp;

	if(!src || !delim || num < 0)
	{
		goto ERROR;
	}

	dst = strarray_create();

	if(!dst)
	{
		goto ERROR;
	}

	sp = (char *)src;

	while(1)
	{
		ep = strstr(sp, delim);

		if(!ep || (num && strarray_len(dst) >= num))
		{
			strarray_add(dst, sp);

			break;
		}

		tmp = strutil_ndup(sp, ep - sp);

		strarray_add(dst, tmp);

		free(tmp);

		sp = ep + strlen(delim);
	}

	return dst;

ERROR:
	return NULL;
}

int
strutil_isempty(const char *src)
{
	return (!src || strlen(src) == 0) ? 1 : 0;
}

int
strutil_isequal(const char *str1, const char *str2)
{
	int   ret;

	if(!str1 || !str2)
	{
		goto ERROR;
	}

	ret = strcmp(str1, str2);

	return ret ? 0 : 1;

ERROR:

	return 0;
}

int
strutil_iscaseequal(const char *str1, const char *str2)
{
	int   ret;

	if(!str1 || !str2)
	{
		goto ERROR;
	}

	ret = strcasecmp(str1, str2);

	return ret ? 0 : 1;

ERROR:

	return 0;
}

int
strutil_isdigit(const char *src)
{
	int len, i;

	if(!src)
	{
		goto ERROR;
	}

	len = strlen(src);

	for(i = 0; i < len; i++)
	{
		if(!isdigit(src[i]))
		{
			break;
		}
	}

	return (len == 0 || i < len) ? 0 : 1;

ERROR:
	return 0;
}

int
strutil_isspace(const char *src)
{
	int len, i;

	if(!src)
	{
		goto ERROR;
	}

	len = strlen(src);

	for(i = 0; i < len; i++)
	{
		if(!isspace(src[i]))
		{
			break;
		}
	}

	return (len == 0 || i < len) ? 0 : 1;

ERROR:
	return 0;
}

int
strutil_isrange(const char *src, int min, int max)
{
	int value;

	if(!src)
	{
		goto ERROR;
	}

	value = strtol(src, NULL, 10);

	return (value < min || value > max) ? 0 : 1;

ERROR:
	return 0;
}

int
strutil_tailis(const char *src, const char *str)
{
	int ret;

	if(!src || !str || strlen(src) < strlen(str))
	{
		goto ERROR;
	}

	ret = strcmp(src + (strlen(src) - strlen(str)), str);

	return ret ? 0 : 1;

ERROR:
	return 0;
}

char *
strutil_take_crlf(char *str)
{
	char *ptr;
	int len;

	ptr = str;

	while(1)
	{
		ptr = strstr(ptr, "\r\n");

		if(!ptr)
		{
			break;
		}

		len = strlen(ptr + 2);

		memmove(ptr, ptr + 2, len);

		ptr[len] = '\0';
	}

	return str;
}

int
strutil_headwith(const char *src, const char *str)
{
	int ret;

	if(!src || !str || strlen(src) < strlen(str))
	{
		goto ERROR;
	}

	ret = strncmp(src, str, strlen(str));

	return ret ? 0 : 1;

ERROR:
	return 0;
}

int
strutil_tailwith(const char *src, const char *str)
{
	int ret;

	if(!src || !str || strlen(src) < strlen(str))
	{
		goto ERROR;
	}

	ret = strcmp(src + (strlen(src) - strlen(str)), str);

	return ret ? 0 : 1;

ERROR:
	return 0;
}

int
strutil_caseheadwith(const char *src, const char *str)
{
	int ret;

	if(!src || !str || strlen(src) < strlen(str))
	{
		goto ERROR;
	}

	ret = strncasecmp(src, str, strlen(str));

	return ret ? 0 : 1;

ERROR:
	return 0;
}

int
strutil_casetailwith(const char *src, const char *str)
{
	int ret;

	if(!src || !str || strlen(src) < strlen(str))
	{
		goto ERROR;
	}

	ret = strcasecmp(src + (strlen(src) - strlen(str)), str);

	return ret ? 0 : 1;

ERROR:
	return 0;
}

char *
strutil_dequote(char *src, const char *squote, const char *equote)
{
	char *sp, *ep;

	sp = src;

	ep = sp + strlen(sp);

	if(!src)
	{
		goto ERROR;
	}

	if(squote)
	{
		if(!strutil_headwith(sp, squote))
		{
			goto ERROR;
		}

		sp += strlen(squote);
	}

	if(equote)
	{
		if(!strutil_tailwith(sp, equote))
		{
			goto ERROR;
		}

		ep -= strlen(equote);
	}

	memmove(src, sp, ep - sp);

	src[ep - sp] = '\0';

	return src;

ERROR:
	return src;
}
