/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdlib.h>
#include <unistd.h>
#include "request.h"
#include "strutil.h"
#include "debug.h"

#define REQUEST_UPLOAD_PATH_TEMPLATE "/tmp/yacafe.upload.XXXXXX"

struct request
{
	list_t *items;
	list_t *upload_items;
};

struct request_item
{
	char *name;
	char *value;
};

struct request_upload_item
{
	char *name;
	char *path;
};

static request_item_t *
request_item_create(void);
static void
request_item_destroy(request_item_t *item);
static request_upload_item_t *
request_upload_item_create(void);
static void
request_upload_item_destroy(request_upload_item_t *item);

request_t *
request_create(void)
{
	request_t *request;

	request = (request_t *)malloc(sizeof(request_t));

	if(request)
	{
		request->items = list_create();

		request->upload_items = list_create();
	}

	return request;
}

void
request_destroy(request_t *request)
{
	if(request)
	{
		while(list_size(request->items))
		{
			request_item_destroy(list_remove(request->items, 0));
		}

		list_destroy(request->items);

		while(list_size(request->upload_items))
		{
			request_item_destroy(list_remove(request->upload_items, 0));
		}

		list_destroy(request->upload_items);
	}

	free(request);
}

int
request_numberof_items(const request_t *request)
{
	return list_size(request->items);
}

request_item_t *
request_append_item(request_t *request, const char *name, const char *value)
{
        request_item_t *item;

        if(!request || strutil_isempty(name))
        {
                goto ERROR;
        }

        item = request_item_create();

        item->name = strutil_dup(name);

        item->value = strutil_dup(value);

        return list_add(request->items, item, list_size(request->items));

ERROR:
        return NULL;
}

void
request_remove_item(request_t *request, int index)
{
	if(!request)
	{
		goto ERROR;
	}

	request_item_destroy(list_remove(request->items, index));

ERROR:
	return;
}

void
request_remove_item_by_name(request_t *request, const char *name)
{
	int size, i;
	request_item_t *item;

	if(!request || !name)
	{
		goto ERROR;
	}

	size = list_size(request->items);

	for(i = 0; i < size; i++)
	{
		item = list_get(request->items, i);

		if(strutil_isequal(item->name, name))
		{
			request_item_destroy(list_remove(request->items, i));

			break;
		}
	}

ERROR:
	return;
}

void
request_remove_items_by_name(request_t *request, const char *name)
{
	int size, i;
	request_item_t *item;

	if(!request || !name)
	{
		goto ERROR;
	}

	size = list_size(request->items);

	for(i = 0; i < size; i++)
	{
		item = list_get(request->items, i);

		if(strutil_isequal(item->name, name))
		{
			request_item_destroy(list_remove(request->items, i));
		}
	}

ERROR:
	return;
}

request_item_t *
request_get_item(const request_t *request, int index)
{
	return request ? list_get(request->items, index) : NULL;
}

request_item_t *
request_get_item_by_name(const request_t *request, const char *name)
{
	int size, i;
	request_item_t *item;

	if(!request || !name)
	{
		goto ERROR;
	}

	size = list_size(request->items);

	for(i = 0; i < size; i++)
	{
		item = list_get(request->items, i);

		if(strutil_isequal(item->name, name))
		{
			return item;
		}
	}

ERROR:
	return NULL;
}

list_t *
request_get_items_by_name(const request_t *request, const char *name)
{
	list_t *dst;
	int size, i;
	request_item_t *item;

	if(!request || !name)
	{
		goto ERROR;
	}

	dst = list_create();

	size = list_size(request->items);

	for(i = 0; i < size; i++)
	{
		item = list_get(request->items, i);

		if(strutil_isequal(item->name, name))
		{
			list_add(dst, item, list_size(dst));
		}
	}

	return dst;

ERROR:
	return NULL;
}

int
request_numberof_upload_items(const request_t *request)
{
	return list_size(request->upload_items);
}

request_upload_item_t *
request_append_upload_item(request_t *request, const char *name)
{
        request_upload_item_t *item;

        if(!request || strutil_isempty(name))
        {
                goto ERROR;
        }

        item = request_upload_item_create();

        item->name = strutil_dup(name);

        item->path = strutil_dup(REQUEST_UPLOAD_PATH_TEMPLATE);

	close(mkstemp(item->path));

        return list_add(request->upload_items, item, list_size(request->upload_items));

ERROR:
        return NULL;
}

void
request_remove_upload_item(request_t *request, int index)
{
	if(request)
	{
		goto ERROR;
	}

	request_upload_item_destroy(list_remove(request->upload_items, index));

ERROR:
	return;
}

void
request_remove_upload_item_by_name(request_t *request, const char *name)
{
	int size, i;
	request_upload_item_t *item;

	if(!request || strutil_isempty(name))
	{
		goto ERROR;
	}

	size = list_size(request->upload_items);

	for(i = 0; i < size; i++)
	{
		item = list_get(request->upload_items, i);

		if(strutil_isequal(item->name, name))
		{
			request_upload_item_destroy(list_remove(request->upload_items, i));

			break;
		}
	}

ERROR:
	return;
}

void
request_remove_upload_items_by_name(request_t *request, const char *name)
{
	int size, i;
	request_upload_item_t *item;

	if(!request || strutil_isempty(name))
	{
		goto ERROR;
	}

	size = list_size(request->upload_items);

	for(i = 0; i < size; i++)
	{
		item = list_get(request->upload_items, i);

		if(strutil_isequal(item->name, name))
		{
			request_upload_item_destroy(list_remove(request->upload_items, i));
		}
	}

ERROR:
	return;
}

request_upload_item_t *
request_get_upload_item(const request_t *request, int index)
{
	return request ? list_get(request->upload_items, index) : NULL;
}

request_upload_item_t *
request_get_upload_item_by_name(const request_t *request, const char *name)
{
	int size, i;
	request_upload_item_t *item;

	if(!request || strutil_isempty(name))
	{
		goto ERROR;
	}

	size = list_size(request->upload_items);

	for(i = 0; i < size; i++)
	{
		item = list_get(request->upload_items, i);

		if(strutil_isequal(item->name, name))
		{
			return item;
		}
	}

ERROR:
	return NULL;
}

list_t *
request_get_upload_items_by_name(const request_t *request, const char *name)
{
	list_t *dst;
	int size, i;
	request_upload_item_t *item;

	if(!request || strutil_isempty(name))
	{
		goto ERROR;
	}

	dst = list_create();

	size = list_size(request->upload_items);

	for(i = 0; i < size; i++)
	{
		item = list_get(request->upload_items, i);

		if(strutil_isequal(item->name, name))
		{
			list_add(dst, item, list_size(dst));
		}
	}

	return dst;

ERROR:
	return NULL;
}

static request_item_t *
request_item_create(void)
{
	request_item_t *item;

	item = (request_item_t *)malloc(sizeof(request_item_t));

	if(item)
	{
		item->name = NULL;

		item->value = NULL;
	}

	return item;
}

static void
request_item_destroy(request_item_t *item)
{
	if(item)
	{
		free(item->name);

		free(item->value);
	}

	free(item);
}

char *
request_item_get_name(const request_item_t *item)
{
	return item ? item->name : NULL;
}

char *
request_item_get_value(const request_item_t *item)
{
	return item ? item->value : NULL;
}

char *
request_item_set_value(request_item_t *item, const char *value)
{
	return item ? strutil_exchange(&item->value, value) : NULL;
}

static request_upload_item_t *
request_upload_item_create(void)
{
	request_upload_item_t *item;

	item = (request_upload_item_t *)malloc(sizeof(request_upload_item_t));

	if(item)
	{
		item->name = NULL;

		item->path = NULL;
	}

	return item;
}

static void
request_upload_item_destroy(request_upload_item_t *item)
{
	if(item)
	{
		free(item->name);

		unlink(item->path);

		free(item->path);
	}

	free(item);
}

char *
request_upload_item_get_name(const request_upload_item_t *item)
{
	return item ? item->name : NULL;
}

char *
request_upload_item_get_path(const request_upload_item_t *item)
{
	return item ? item->path : NULL;
}

char *
request_upload_item_set_path(request_upload_item_t *item, const char *path)
{
	return item ? strutil_exchange(&item->path, path) : NULL;
}
