/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdlib.h>
#include <unistd.h>
#include "response.h"
#include "strutil.h"
#include "debug.h"

#define REQUEST_UPLOAD_PATH_TEMPLATE "/tmp/yacafe.upload.XXXXXX"

struct response
{
	list_t *nodes;
};

struct response_node
{
	char *name;
	list_t *nodes;
	list_t *properties;
};

struct response_node_property
{
	char *name;
	char *value;
};

static response_node_t *
response_node_create(void);
static void
response_node_destroy(response_node_t *node);
static response_node_property_t *
response_node_property_create(void);
static void
response_node_property_destroy(response_node_property_t *property);

response_t *
response_create(void)
{
	response_t *response;

	response = (response_t *)malloc(sizeof(response_t));

	if(response)
	{
		response->nodes = list_create();
	}

	return response;
}

void
response_destroy(response_t *response)
{
	if(response)
	{
		while(list_size(response->nodes))
		{
			response_node_destroy(list_remove(response->nodes, 0));
		}

		list_destroy(response->nodes);
	}

	free(response);
}

int
response_numberof_nodes(const response_t *response)
{
	return list_size(response->nodes);
}

response_node_t *
response_append_node(response_t *response, const char *name)
{
        response_node_t *node;

        if(!response || strutil_isempty(name))
        {
                goto ERROR;
        }

        node = response_node_create();

        node->name = strutil_dup(name);

        return list_add(response->nodes, node, list_size(response->nodes));

ERROR:
        return NULL;
}

void
response_remove_node(response_t *response, int index)
{
	if(!response)
	{
		goto ERROR;
	}

	response_node_destroy(list_remove(response->nodes, index));

ERROR:
	return;
}

void
response_remove_node_by_name(response_t *response, const char *name)
{
	int size, i;
	response_node_t *node;

	if(!response || !name)
	{
		goto ERROR;
	}

	size = list_size(response->nodes);

	for(i = 0; i < size; i++)
	{
		node = list_get(response->nodes, i);

		if(strutil_isequal(node->name, name))
		{
			response_node_destroy(list_remove(response->nodes, i));

			break;
		}
	}

ERROR:
	return;
}

void
response_remove_nodes_by_name(response_t *response, const char *name)
{
	int size, i;
	response_node_t *node;

	if(!response || !name)
	{
		goto ERROR;
	}

	size = list_size(response->nodes);

	for(i = 0; i < size; i++)
	{
		node = list_get(response->nodes, i);

		if(strutil_isequal(node->name, name))
		{
			response_node_destroy(list_remove(response->nodes, i));
		}
	}

ERROR:
	return;
}

response_node_t *
response_get_node(const response_t *response, int index)
{
	return response ? list_get(response->nodes, index) : NULL;
}

response_node_t *
response_get_node_by_name(const response_t *response, const char *name)
{
	int size, i;
	response_node_t *node;

	if(!response || !name)
	{
		goto ERROR;
	}

	size = list_size(response->nodes);

	for(i = 0; i < size; i++)
	{
		node = list_get(response->nodes, i);

		if(strutil_isequal(node->name, name))
		{
			return node;
		}
	}

ERROR:
	return NULL;
}

list_t *
response_get_nodes_by_name(const response_t *response, const char *name)
{
	list_t *dst;
	int size, i;
	response_node_t *node;

	if(!response || !name)
	{
		goto ERROR;
	}

	dst = list_create();

	size = list_size(response->nodes);

	for(i = 0; i < size; i++)
	{
		node = list_get(response->nodes, i);

		if(strutil_isequal(node->name, name))
		{
			list_add(dst, node, list_size(dst));
		}
	}

	return dst;

ERROR:
	return NULL;
}

static response_node_t *
response_node_create(void)
{
	response_node_t *node;

	node = (response_node_t *)malloc(sizeof(response_node_t));

	if(node)
	{
		node->name = NULL;

		node->nodes = list_create();

		node->properties = list_create();
	}

	return node;
}

static void
response_node_destroy(response_node_t *node)
{
	if(node)
	{
		free(node->name);

		while(list_size(node->nodes))
		{
			response_node_destroy(list_remove(node->nodes, 0));
		}

		list_destroy(node->nodes);

		while(list_size(node->properties))
		{
			response_node_property_destroy(list_remove(node->properties, 0));
		}

		list_destroy(node->properties);
	}

	free(node);
}

int
response_node_numberof_nodes(const response_node_t *node)
{
	return node ? list_size(node->nodes) : -1;
}

int
response_node_numberof_properties(const response_node_t *node)
{
	return node ? list_size(node->properties) : -1;
}

char *
response_node_get_name(const response_node_t *node)
{
	return node ? node->name : NULL;
}

char *
response_node_set_name(response_node_t *node, const char *name)
{
	return node ? strutil_exchange(&node->name, name) : NULL;
}

response_node_t *
response_node_append_node(response_node_t *node, const char *name)
{
        response_node_t *dst;

        if(!node || strutil_isempty(name))
        {
                goto ERROR;
        }

        dst = response_node_create();

        dst->name = strutil_dup(name);

        return list_add(node->nodes, dst, list_size(node->nodes));

ERROR:
        return NULL;
}

response_node_property_t *
response_node_append_property(response_node_t *node, const char *name, const char *value)
{
        response_node_property_t *dst;

        if(!node || strutil_isempty(name))
        {
                goto ERROR;
        }

        dst = response_node_property_create();

        dst->name = strutil_dup(name);
        dst->value = strutil_dup(value);

        return list_add(node->properties, dst, list_size(node->properties));

ERROR:
        return NULL;
}

void
response_node_remove_node(response_node_t *node, int index)
{
	if(!node)
	{
		goto ERROR;
	}

	response_node_destroy(list_remove(node->nodes, index));

ERROR:
	return;
}

void
response_node_remove_property(response_node_t *node, int index)
{
	if(!node)
	{
		goto ERROR;
	}

	response_node_property_destroy(list_remove(node->properties, index));

ERROR:
	return;
}

void
response_node_remove_node_by_name(response_node_t *node, const char *name)
{
	int size, i;
	response_node_t *tmp;

	if(!node || !name)
	{
		goto ERROR;
	}

	size = list_size(node->nodes);

	for(i = 0; i < size; i++)
	{
		tmp = list_get(node->nodes, i);

		if(strutil_isequal(tmp->name, name))
		{
			response_node_destroy(list_remove(node->nodes, i));

			break;
		}
	}

ERROR:
	return;
}

void
response_node_remove_property_by_name(response_node_t *node, const char *name)
{
	int size, i;
	response_node_property_t *tmp;

	if(!node || !name)
	{
		goto ERROR;
	}

	size = list_size(node->properties);

	for(i = 0; i < size; i++)
	{
		tmp = list_get(node->properties, i);

		if(strutil_isequal(tmp->name, name))
		{
			response_node_property_destroy(list_remove(node->properties, i));

			break;
		}
	}

ERROR:
	return;
}

void
response_node_remove_nodes_by_name(response_node_t *node, const char *name)
{
	int size, i;
	response_node_t *tmp;

	if(!node || !name)
	{
		goto ERROR;
	}

	size = list_size(node->nodes);

	for(i = 0; i < size; i++)
	{
		tmp = list_get(node->nodes, i);

		if(strutil_isequal(tmp->name, name))
		{
			response_node_destroy(list_remove(node->nodes, i));
		}
	}

ERROR:
	return;
}

void
response_node_remove_properties_by_name(response_node_t *node, const char *name)
{
	int size, i;
	response_node_property_t *tmp;

	if(!node || !name)
	{
		goto ERROR;
	}

	size = list_size(node->properties);

	for(i = 0; i < size; i++)
	{
		tmp = list_get(node->properties, i);

		if(strutil_isequal(tmp->name, name))
		{
			response_node_property_destroy(list_remove(node->properties, i));
		}
	}

ERROR:
	return;
}

response_node_t *
response_node_get_node(response_node_t *node, int index)
{
	return node ? list_get(node->nodes, index) : NULL;
}

response_node_property_t *
response_node_get_property(response_node_t *node, int index)
{
	return node ? list_get(node->properties, index) : NULL;
}

response_node_t *
response_node_get_node_by_name(response_node_t *node, const char *name)
{
	int size, i;
	response_node_t *dst;

	if(!node || !name)
	{
		goto ERROR;
	}

	size = list_size(node->nodes);

	for(i = 0; i < size; i++)
	{
		dst = list_get(node->nodes, i);

		if(strutil_isequal(dst->name, name))
		{
			return dst;
		}
	}

ERROR:
	return NULL;
}

response_node_property_t *
response_node_get_property_by_name(response_node_t *node, const char *name)
{
	int size, i;
	response_node_property_t *dst;

	if(!node || !name)
	{
		goto ERROR;
	}

	size = list_size(node->properties);

	for(i = 0; i < size; i++)
	{
		dst = list_get(node->properties, i);

		if(strutil_isequal(dst->name, name))
		{
			return dst;
		}
	}

ERROR:
	return NULL;
}

list_t *
response_node_get_nodes_by_name(response_node_t *node, const char *name)
{
	list_t *dst;
	int size, i;
	response_node_t *tmp;

	if(!node || !name)
	{
		goto ERROR;
	}

	dst = list_create();

	size = list_size(node->nodes);

	for(i = 0; i < size; i++)
	{
		tmp = list_get(node->nodes, i);

		if(strutil_isequal(tmp->name, name))
		{
			list_add(dst, tmp, list_size(dst));
		}
	}

	return dst;

ERROR:
	return NULL;
}

list_t *
response_node_get_properties_by_name(response_node_t *node, const char *name)
{
	list_t *dst;
	int size, i;
	response_node_property_t *tmp;

	if(!node || !name)
	{
		goto ERROR;
	}

	dst = list_create();

	size = list_size(node->properties);

	for(i = 0; i < size; i++)
	{
		tmp = list_get(node->properties, i);

		if(strutil_isequal(tmp->name, name))
		{
			list_add(dst, tmp, list_size(dst));
		}
	}

	return dst;

ERROR:
	return NULL;
}

static response_node_property_t *
response_node_property_create(void)
{
	response_node_property_t *property;

	property = (response_node_property_t *)malloc(sizeof(response_node_property_t));

	if(property)
	{
		property->name = NULL;

		property->value = NULL;
	}

	return property;
}

static void
response_node_property_destroy(response_node_property_t *property)
{
	if(property)
	{
		free(property->name);

		free(property->value);
	}

	free(property);
}

char *
response_node_property_get_name(const response_node_property_t *property)
{
	return property ? property->name : NULL;
}

char *
response_node_property_get_value(const response_node_property_t *property)
{
	return property ? property->value : NULL;
}

char *
response_node_property_set_name(response_node_property_t *property, const char *name)
{
	return property ? strutil_exchange(&property->name, name) : NULL;
}

char *
response_node_property_set_value(response_node_property_t *property, const char *value)
{
	return property ? strutil_exchange(&property->value, value) : NULL;
}
