/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "session.h"
#include "semlock.h"
#include "logger.h"
#include "debug.h"

#define SESSION_KEY 20080215
#define SESSION_MAX 1000
#define SESSION_TIMEOUT 300

typedef struct session_info session_info_t;

struct session
{
	void *shm;
	char id[128];
	session_info_t *info;
};

struct session_info
{
	unsigned long id;
	unsigned long expire;
};

static session_info_t *
session_info_get(session_t *session, unsigned long id);
static session_info_t *
session_info_add(session_t *session);
static unsigned long
session_generate_id(void);
static unsigned long
session_generate_timestamp(int offset);

extern logger_t *glogger;

/*
int
main(int argc, char *argv)
{
	session_t *session;
	session_info_t *ptr;
	char *sessionid;
	int i, j;

	session = session_create();

	if(!session)
	{
		printf("session_create(): error\n");

		goto ERROR;
	}

	ptr = session->shm;

	for(i = 0; i < SESSION_MAX; i++)
	{
		ptr[i].id = 0;
		ptr[i].expire = 0;
		printf("deleted[%03d]: %s\n", i, sessionid);
		usleep(100);
	}

	for(j = 0; j < 1200; j++)
	{
		sessionid = session_open(session, NULL);

		printf("sessionid[%03d]: %s\n", j, sessionid);

		usleep(100);
	}

	sleep(3);

	for(j = 0; j < 1200; j++)
	{
		sessionid = session_open(session, NULL);

		printf("sessionid[%03d]: %s\n", j, sessionid);

		usleep(100);
	}

	sleep(10);

	for(j = 0; j < 1200; j++)
	{
		sessionid = session_open(session, NULL);

		printf("sessionid[%03d]: %s\n", j, sessionid);

		usleep(100);
	}	

	session_destroy(session);

	return  0;

ERROR:
	session_destroy(session);

	return -1;
}
*/

session_t *
session_create(void)
{
	session_t *session = NULL;

	session = (session_t *)malloc(sizeof(session_t));

	if(session)
	{
		int shmid;

		shmid = shmget((key_t)SESSION_KEY, sizeof(session_info_t) * SESSION_MAX, IPC_CREAT | 0666);

		if(shmid == -1)
		{
			goto ERROR;
		}

		session->shm = shmat(shmid, NULL, 0);

		if(session->shm == (void *)-1)
		{
			goto ERROR;
		}

		memset(session->id, 0x00, sizeof(session->id));

		session->info = NULL;
	}

	return session;

ERROR:
	free(session);

	return NULL;
}

void
session_destroy(session_t *session)
{
	if(session)
	{
		shmdt(session->shm);
	}

	free(session);
}

char *
session_open(session_t *session, const char *id)
{
	semlock_t *semlock;

	logger_entry_function(glogger);

	logger_printf(glogger, LOGGER_DEBUG, "id: %s", id);

	if(!session || !session->shm)
	{
		goto ERROR;
	}

	semlock = semlock_open(SESSION_KEY);

	if(!semlock)
	{
		goto ERROR;
	}

	session->info = id ? session_info_get(session, strtoul(id, NULL, 10)) : session_info_add(session);

	if(!session->info)
	{
		goto ERROR;
	}

	semlock_close(semlock);

	sprintf(session->id, "%lu", session->info->id);

	logger_printf(glogger, LOGGER_DEBUG, "session->id: %s ", session->id);

	logger_leave_function(glogger);

	return session->id;

ERROR:
	semlock_close(semlock);

	logger_printf(glogger, LOGGER_DEBUG, "session open error");

	logger_leave_function(glogger);

	return NULL;
}

void
session_close(session_t *session)
{
	logger_entry_function(glogger);

	if(!session || !session->info)
	{
		goto ERROR;
	}

	memset(session->id, 0x00, sizeof(session->id));

	session->info->id = 0;
	session->info->expire = 0;
	session->info = NULL;

ERROR:
	logger_leave_function(glogger);

	return;
}

int
session_isvalid(const session_t *session)
{
	session_info_t *info;

	logger_entry_function(glogger);

	if(!session || !session->info)
	{
		goto ERROR;
	}

	logger_printf(glogger, LOGGER_DEBUG, "sessionid: %s", session->id);

	info = session->info;

	if(info->id == 0 || info->expire <= session_generate_timestamp(0))
	{
		goto ERROR;
	}

	logger_printf(glogger, LOGGER_DEBUG, "this session is valid");

	logger_leave_function(glogger);

	return 1;

ERROR:
	logger_printf(glogger, LOGGER_DEBUG, "this session is invalid");

	logger_leave_function(glogger);

	return 0;
}

char *
session_get_id(const session_t *session)
{
	return session ? (char *)session->id : NULL;
}

static session_info_t *
session_info_get(session_t *session, unsigned long id)
{
	session_info_t *ptr;
	int i;

	logger_entry_function(glogger);

	if(!session || !session->shm)
	{
		goto ERROR;
	}

	ptr = session->shm;

	for(i = 0; i < SESSION_MAX; i++)
	{
		if(ptr[i].id == id && ptr[i].expire > session_generate_timestamp(0))
		{
			ptr[i].expire = session_generate_timestamp(SESSION_TIMEOUT);

			logger_printf(glogger, LOGGER_DEBUG, "session found");

			logger_leave_function(glogger);

			return ptr + i;
		}
	}

	logger_printf(glogger, LOGGER_DEBUG, "session not found");

ERROR:
	logger_leave_function(glogger);

	return NULL;
}

static session_info_t *
session_info_add(session_t *session)
{
	session_info_t *ptr;
	int i;

	if(!session || !session->shm)
	{
		goto ERROR;
	}

	ptr = session->shm;

	for(i = 0; i < SESSION_MAX; i++)
	{
		if(ptr[i].id == 0 || ptr[i].expire <= session_generate_timestamp(0))
		{
			ptr[i].id = session_generate_id();

			ptr[i].expire = session_generate_timestamp(SESSION_TIMEOUT);

			return ptr + i;
		}
	}

ERROR:
	return NULL;
}

static unsigned long
session_generate_id(void)
{
int           fd;
unsigned long dst;

	fd = open("/dev/urandom", O_RDONLY);

	if(fd < 0)
	{
		goto ERROR;
	}

	read(fd, &dst, sizeof(dst));

	close(fd);

	return dst;

ERROR:
	return 0;
}

static unsigned long
session_generate_timestamp(int offset)
{
struct timeval tv;

	gettimeofday(&tv, NULL);

	return tv.tv_sec + offset;
}
