/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include "template.h"
#include "list.h"
#include "pairarray.h"
#include "strarray.h"
#include "strutil.h"
#include "stack.h"
#include "debug.h"

#define BTAG "{{"
#define ETAG "}}"

struct foreach_info
{
	list_t *nodes;
	char *position;
	int count;
};

typedef struct foreach_info foreach_info_t;

static void
template_parse_core(FILE *fp, const char *data, const response_t *response, const session_t *session);
static pairarray_t *
template_parse_tag(const char *tag);
static int
template_parse_tag_one(pairarray_t *array, const char *tag);
static int
template_isactive(mystack_t *stack);
static list_t *
template_search_nodes(const response_t *response, const char *target, const mystack_t *stack);
static response_node_t *
template_search_node(const response_t *response, const char *target, const mystack_t *stack);
static char *
template_search_text(const char *src, int *idx);

int
template_parse(FILE *fp, const char *file, const response_t *response, const session_t *session)
{
	int fd;
	struct stat st;
	char *ptr;

	fd = open(file, O_RDONLY);

	if(fd < 0)
	{
		goto ERROR;
	}

	fstat(fd, &st);

	ptr = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);

	if(!ptr)
	{
		goto ERROR;
	}

	template_parse_core(fp, ptr, response, session);

	munmap(ptr, st.st_size);

	close(fd);

	return  0;

ERROR:
	if(fd >= 0)
	{
		close(fd);
	}

	return -1;
}

static void
template_parse_core(FILE *fp, const char *src, const response_t *response, const session_t *session)
{
	char *sp, *ep, *tmp, *type;
	pairarray_t *tag;
	mystack_t *stack;

	ep = (char *)src;

	stack = stack_create();

	while(1)
	{
		sp = strstr(ep, BTAG);

		if(!sp)
		{
			if(template_isactive(stack))
			{
				fwrite(ep, sizeof(char), strlen(ep), fp);
			}

			break;
		}

		if(template_isactive(stack))
		{
			fwrite(ep, sizeof(char), sp - ep, fp);
		}

		sp += strlen(BTAG);

		ep = strstr(sp, ETAG);

		if(!ep)
		{
			break;
		}

		tmp = strutil_ndup(sp, ep - sp);

		tag = template_parse_tag(tmp);

		type = pairarray_get_value_by_name(tag, "type");

		if(strutil_isequal(type, "property"))
		{
			response_node_t *target;
			char *property;

			if(template_isactive(stack))
			{
				target = template_search_node(response, pairarray_get_value_by_name(tag, "target"), stack);

				property = response_node_property_get_value(response_node_get_property_by_name(target, pairarray_get_value_by_name(tag, "property")));

				fprintf(fp, "%s", property ? property : "");
			}
		}
		else if(strutil_isequal(type, "session"))
		{
			char *target, *sessionid;

			if(template_isactive(stack))
			{
				target = pairarray_get_value_by_name(tag, "target");

				if(strutil_isequal(target, "id"))
				{
					sessionid = session_get_id(session);

					fprintf(fp, "%s", sessionid ? sessionid : "");
				}
			}
		}
		else if(strutil_isequal(type, "foreach"))
		{
			foreach_info_t *info;

			info = (foreach_info_t *)malloc(sizeof(foreach_info_t));

			info->nodes = template_search_nodes(response, pairarray_get_value_by_name(tag, "target"), stack);

			info->position = ep;

			info->count = template_isactive(stack) ? list_size(info->nodes) : 0;

			stack_push(stack, info);
		}
		else if(strutil_isequal(type, "end"))
		{
			foreach_info_t *info;

			info = stack_peek(stack);

			if(info)
			{
				if(info->count > 0)
				{
					ep = info->position;

					info->count -= 1;
				}
				else
				{
					foreach_info_t *info;

					info = stack_pop(stack);

					if(info)
					{
						list_destroy(info->nodes);
					}

					free(info);
				}
			}
		}
		else if(strutil_isequal(type, "include"))
		{
			if(template_isactive(stack))
			{
				template_parse(fp, pairarray_get_value_by_name(tag, "target"), response, session);
			}
		}

		pairarray_destroy(tag);

		free(tmp);

		ep += strlen(ETAG);
	}

	stack_destroy(stack);
}

static pairarray_t *
template_parse_tag(const char *tag)
{
	pairarray_t *dst;
	strarray_t *tmp;
	int len, i;

	if(!tag)
	{
		goto ERROR;
	}

	dst = pairarray_create();

	tmp = strutil_split(tag, ";", 0);

	len = strarray_len(tmp);

	for(i = 0; i < len; i++)
	{
		template_parse_tag_one(dst, strarray_get(tmp, i));
	}

	strarray_destroy(tmp);

	return dst;

ERROR:
	return NULL;
}

static int
template_parse_tag_one(pairarray_t *array, const char *tag)
{
	strarray_t *tmp;
	char *name, *value;

	if(!array || !tag)
	{
		goto ERROR;
	}

	tmp = strutil_split(tag, "=", 1);

	name  = strutil_trim(strarray_get(tmp, 0));

	value = strutil_trim(strarray_get(tmp, 1));

	pairarray_set_value_by_name(array, name, value);

	free(name); free(value);

	strarray_destroy(tmp);

	return  0;

ERROR:
	return -1;
}

static int
template_isactive(mystack_t *stack)
{
	foreach_info_t *info;

	info = stack_peek(stack);

	if(info && info->count <= 0)
	{
		return 0;
	}

	return 1;
}

static list_t *
template_search_nodes(const response_t *response, const char *target, const mystack_t *stack)
{
	strarray_t *path;
	int len, i, idx;
	char *tmp, *name;
	list_t *dst = NULL;
	response_node_t *node = NULL;

	path = strutil_split(target, "/", 0);

	len = strarray_len(path);

	for(i = 0; i < len; i++)
	{
		tmp = strarray_get(path, i);

		if(strutil_isempty(tmp))
		{
			continue;
		}

		name = template_search_text(tmp, &idx);

		if(!name)
		{
			break;
		}

		if(!node && strutil_isequal(name, "$$"))
		{
			foreach_info_t *info;

			info = stack_peek(stack);

			if(info)
			{
				node = list_get(info->nodes, list_size(info->nodes) - info->count);
			}
		}
		else
		{
			dst = node ? response_node_get_nodes_by_name(node, name) : response_get_nodes_by_name(response, name);

			if(!dst)
			{
				free(name);

				break;
			}

			node = list_get(dst, idx);
		}

		free(name);

		if(!node && i < (len - 1))
		{
			break;
		}
	}

	strarray_destroy(path);

	if(i < len)
	{
		goto ERROR;
	}

	return dst;

ERROR:
	return NULL;
}

static response_node_t *
template_search_node(const response_t *response, const char *target, const mystack_t *stack)
{
	strarray_t *path;
	int len, i, idx;
	char *tmp, *name;
	list_t *list = NULL;
	response_node_t *node = NULL;

	path = strutil_split(target, "/", 0);

	len = strarray_len(path);

	for(i = 0; i < len; i++)
	{
		tmp = strarray_get(path, i);

		if(strutil_isempty(tmp))
		{
			continue;
		}

		name = template_search_text(tmp, &idx);

		if(!name)
		{
			break;
		}

		if(!node && strutil_isequal(name, "$$"))
		{
			foreach_info_t *info;

			info = stack_peek(stack);

			if(info)
			{
				node = list_get(info->nodes, list_size(info->nodes) - info->count);
			}
		}
		else
		{
			list = node ? response_node_get_nodes_by_name(node, name) : response_get_nodes_by_name(response, name);

			if(!list)
			{
				free(name);

				break;
			}

			node = list_get(list, idx);

			list_destroy(list);
		}

		free(name);

		if(!node)
		{
			break;
		}
	}

	strarray_destroy(path);

	if(i < len)
	{
		goto ERROR;
	}

	return node;

ERROR:
	return NULL;
}

static char *
template_search_text(const char *src, int *idx)
{
	char *sp, *ep, *tmp, *dst;

	if(!src || !idx)
	{
		goto ERROR;
	}

	sp = strstr(src, "[");

	if(sp)
	{
		ep  = strstr(sp, "]");

		if(!ep)
		{
			goto ERROR;
		}

		tmp = strutil_ndup(sp + 1, ep - (sp + 1));

		if(!strutil_isdigit(tmp))
		{
			free(tmp);

			goto ERROR;
		}

		*idx = strtol(tmp, NULL, 10);

		free(tmp);

		dst = strutil_ndup(src, sp - src);
	}
	else
	{
		*idx = 0;

		dst = strutil_dup(src);
	}

	return dst;

ERROR:
	return NULL;
}
