/*
 * Copyright 2008 The Yacafe Project. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE YACAFE PROJECT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE YACAFE PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the Yacafe Project.
 */

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include "yacafe.h"
#include "cgi.h"
#include "request.h"
#include "response.h"
#include "session.h"
#include "template.h"
#include "conf.h"
#include "logger.h"
#include "strutil.h"
#include "xstring.h"
#include "debug.h"

#define YACAFE_FAILURE_SESSION_TITLE    "å󥨥顼"
#define YACAFE_FAILURE_SESSION_MESSAGE  "Υå̵Ǥ̤ޤ"
#define YACAFE_FAILURE_SESSION_BUTTON   "λ"
#define YACAFE_FAILURE_SESSION_SCRIPT   "location.replace('.')"

static int
yacafe_core(const conf_t *conf, const char *name, request_t *request, session_t *session);
static char *
yacafe_action(const conf_t *conf, const char *name, request_t *request, response_t *response, session_t *session);
static int
yacafe_forward(const conf_t *conf, const char *name, const char *status, request_t *request, response_t *response, session_t *session);
static int
yacafe_view(const conf_t *conf, const char *name, const response_t *response, const session_t *session);
static void
yacafe_error(const char *title, const char *summary, const char *detail);

extern logger_t *glogger;

int
main(int argc, char *argv[])
{
	conf_t *conf;
	request_t *request = NULL;
	response_t *response = NULL;
	session_t *session = NULL;
	char *sessionid, *app;
	int err;

	glogger = logger_create();

	logger_setopt_name(glogger, "yacafe");

	logger_setopt_file(glogger, "yacafe.log");

	logger_open(glogger);

	logger_printf(glogger, LOGGER_INFO, "------------------------------------------------------------------------------");
	logger_printf(glogger, LOGGER_INFO, " yacafe version %s", VERSION);
	logger_printf(glogger, LOGGER_INFO, "------------------------------------------------------------------------------");

	conf = conf_create();

	if(!conf)
	{
		goto ERROR;
	}

	err = conf_load(conf, CONF_FILE);

	if(err)
	{
		yacafe_error("yacafe error", "Internal Server Error", "The configuration file cannot be read.");

		goto ERROR;
	}

	request = cgi_request_get();

	if(!request)
	{
		yacafe_error("yacafe error", "Internal Server Error", "The request data was not able to be acquired.");

		goto ERROR;
	}

	session = session_create();

	if(!session)
	{
		yacafe_error("yacafe error", "Internal Server Error", "The session object was not able to be created.");

		goto ERROR;
	}

	sessionid = request_item_get_value(request_get_item_by_name(request, "sessionid"));

	if(sessionid)
	{
		session_open(session, sessionid);
	}

	app = request_item_get_value(request_get_item_by_name(request, "app"));

	if(strutil_isempty(app))
	{
		app = "login";
	}

	err = yacafe_core(conf, app, request, session);

	if(err)
	{
		goto ERROR;
	}

	session_destroy(session);

	response_destroy(response);

	request_destroy(request);

	conf_destroy(conf);

	logger_destroy(glogger);

	return  0;

ERROR:
	session_destroy(session);

	response_destroy(response);

	request_destroy(request);

	conf_destroy(conf);

	logger_destroy(glogger);

	return -1;
}

static int
yacafe_core(const conf_t *conf, const char *name, request_t *request, session_t *session)
{
	response_t *response;
	char *status = NULL;
	int err;

	logger_entry_function(glogger);

	response = response_create();

	if(!response)
	{
		goto ERROR;
	}

	status = yacafe_action(conf, name, request, response, session);

	if(strutil_isempty(status))
	{
		goto ERROR;
	}

	err = yacafe_forward(conf, name, status, request, response, session);

	if(err)
	{
		goto ERROR;
	}

	free(status);

	response_destroy(response);

	logger_leave_function(glogger);

	return  0;

ERROR:
	free(status);

	response_destroy(response);

	logger_leave_function(glogger);

	return -1;
}

static char *
yacafe_action(const conf_t *conf, const char *name, request_t *request, response_t *response, session_t *session)
{
	conf_section_t *section;
	void *dl = NULL;
	char *(*action)(request_t *, response_t *, session_t *);
	char *module, *dst, *tmp;

	logger_entry_function(glogger);

	logger_printf(glogger, LOGGER_DEBUG, "name: %s", name);

	section = conf_get_section_by_name(conf, name);

	if(!section)
	{
		yacafe_error("yacafe error", "Not Found", "The requested application was not found.");

		goto ERROR;
	}

	module = conf_section_item_get_value(conf_section_get_item_by_name(section, "module"));

	logger_printf(glogger, LOGGER_DEBUG, "module: %s", module);

	if(strutil_isempty(module))
	{
		yacafe_error("yacafe error", "Internal Server Error", "module not found");

		goto ERROR;;
	}

	dl = dlopen(module, RTLD_NOW);

	if(!dl)
	{
		yacafe_error("yacafe error", "Internal Server Error", dlerror());

		goto ERROR;
	}

	action = dlsym(dl, "action");

	if(!action)
	{
		yacafe_error("yacafe error", "Internal Server Error", dlerror());

		goto ERROR;
	}

	tmp = conf_section_item_get_value(conf_section_get_item_by_name(section, "order"));

	logger_printf(glogger, LOGGER_DEBUG, "order: %s", tmp);

	if(strutil_isequal(tmp, "deny"))
	{
		tmp = request_item_get_value(request_get_item_by_name(request, "app"));

		if(strutil_isequal(tmp, name))
		{
			yacafe_error("yacafe error", "Forbidden", "You don't have permission to access application.");

			goto ERROR;
		}
	}

	tmp = conf_section_item_get_value(conf_section_get_item_by_name(section, "session"));

	logger_printf(glogger, LOGGER_DEBUG, "session: %s", tmp);

	if(strutil_isequal(tmp, "valid") && !session_isvalid(session))
	{
		request_append_item(request, "dialog.title", YACAFE_FAILURE_SESSION_TITLE);
		request_append_item(request, "dialog.message", YACAFE_FAILURE_SESSION_MESSAGE);
		request_append_item(request, "dialog.button", YACAFE_FAILURE_SESSION_BUTTON);
		request_append_item(request, "dialog.script", YACAFE_FAILURE_SESSION_SCRIPT);

		dst = strutil_dup("failure");
	}
	else
	{
		tmp = (*action)(request, response, session);

		dst = strutil_dup(tmp);

		free(tmp);
	}

	dlclose(dl);

	logger_printf(glogger, LOGGER_DEBUG, "dst: %s", dst);

	logger_leave_function(glogger);

	return dst;

ERROR:
	if(dl)
	{
		dlclose(dl);
	}

	logger_leave_function(glogger);

	return NULL;
}

static int
yacafe_forward(const conf_t *conf, const char *name, const char *status, request_t *request, response_t *response, session_t *session)
{
	conf_section_t *section;
	char *tmp, *forward;
	int err;

	logger_entry_function(glogger);

	logger_printf(glogger, LOGGER_DEBUG, "name: %s", name);

	logger_printf(glogger, LOGGER_DEBUG, "status: %s", status);

	section = conf_get_section_by_name(conf, name);

	if(!section)
	{
		goto ERROR;
	}

	tmp = strutil_unit("forward", status, ":");

	forward = conf_section_item_get_value(conf_section_get_item_by_name(section, tmp));

	free(tmp);

	if(strutil_isequal(forward, "default"))
	{
		err = yacafe_view(conf, name, response, session);
	}
	else
	{
		err = yacafe_core(conf, forward, request, session);
	}

	if(err)
	{
		goto ERROR;
	}

	logger_leave_function(glogger);

	return  0;

ERROR:
	logger_leave_function(glogger);

	return -1;
}

static int
yacafe_view(const conf_t *conf, const char *name, const response_t *response, const session_t *session)
{
	conf_section_t *section;
	char *content_type, *response_processor, *param, *template;
	void *dl = NULL;
	int (*execute)(FILE *, const char *, const response_t *, const session_t *);
	int ret;

	logger_entry_function(glogger);

	section = conf_get_section_by_name(conf, name);

	if(!section)
	{
		goto ERROR;
	}

	content_type = conf_section_item_get_value(conf_section_get_item_by_name(section, "content-type"));

	if(!strutil_isempty(content_type))
	{
		printf("Content-type: %s\n\n", content_type);
	}

	response_processor = conf_section_item_get_value(conf_section_get_item_by_name(section, "response_processor"));

	if(!strutil_isempty(response_processor))
	{
		dl = dlopen(response_processor, RTLD_NOW);

		if(!dl)
		{
			goto ERROR;
		}

		execute = dlsym(dl, "execute");

		if(!execute)
		{
			goto ERROR;
		}

		param = conf_section_item_get_value(conf_section_get_item_by_name(section, "response_processor:param"));

		ret = (*execute)(stdout, param, response, session);

		dlclose(dl);
	}
	else
	{
		template = conf_section_item_get_value(conf_section_get_item_by_name(section, "template"));

		ret = template_parse(stdout, template, response, session);
	}

	logger_leave_function(glogger);

	return ret;

ERROR:
	if(dl)
	{
		dlclose(dl);
	}

	logger_leave_function(glogger);

	return -1;
}

static void
yacafe_error(const char *title, const char *summary, const char *detail)
{
	logger_entry_function(glogger);

	logger_printf(glogger, LOGGER_DEBUG, "title: %s", title);
	logger_printf(glogger, LOGGER_DEBUG, "summary: %s", summary);
	logger_printf(glogger, LOGGER_DEBUG, "detail: %s", detail);

	printf("Content-type: text/html\n\n");

	printf("<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n"); 
	printf("<html>\n");
	printf("\t<head>\n");
	printf("\t\t<title>%s</title>\n", title);
	printf("\t</head>\n");
	printf("\t<body>\n");
	printf("\t\t<h1>yacafe error: %s</h1>\n", summary);
	printf("\t\t<p>%s</p>\n", detail);
	printf("\t<body>\n");
	printf("<html>\n");

	logger_leave_function(glogger);
}
